# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from collections import OrderedDict
import inspect
import sys
from typing import Optional

from contrast.utils.decorators import fail_quietly
from contrast_vendor import wrapt

from contrast.agent.policy import patch_manager
from contrast.utils.ignored_modules import (
    ALL_MODULES_TO_IGNORE,
    MODULES_TO_IGNORE_PREFIXES,
)
from contrast.utils.object_utils import get_name

from contrast_vendor.wrapt import function_wrapper
from contrast_vendor import structlog as logging

from contrast.utils.string_utils import ensure_string

logger = logging.getLogger("contrast")


def add_watermark(func):
    """
    Adds a "secret" attribute to patched function for debugging purposes.

    Do not rely on the existence of this attribute in agent source code.
    """
    try:
        func.__contrast__ = True
    except Exception:
        pass
    return func


def wrap_and_watermark(orig_func, wrapper):
    # NOTE: adding a watermark here doesn't make a ton of sense anymore since
    # 1. The function wrapper always has a __wrapped__ attribute, which is a good watermark itself
    # 2. We can't apply the watermark to the wrapped function because in that
    # case it actually gets applied to the underlying object (which we don't want).
    return function_wrapper(add_watermark(wrapper))(orig_func)


def pack_self(instance: Optional[object], args: tuple) -> tuple:
    """Instance can never be None"""
    return args if instance is None else (instance,) + args


def module_is_ignored(module_name):
    return module_name in ALL_MODULES_TO_IGNORE or module_name.startswith(
        MODULES_TO_IGNORE_PREFIXES
    )


def build_and_apply_patch(
    owner,
    attr_name,
    patch_builder,
    builder_args=None,
    owner_name=None,
    force=False,
):
    """
    Builds new patch using given builder and applies it to specified patch location

    :param owner: Module or class where patch will be applied
    :param loc_name: Fully specified name of module or class where patch will apply
    :param attr_name: Name of the method that is being patched/replaced
    :param patch_builder: Callback function used to build new patch
    :param builder_args: A tuple of positional args to be passed to `patch_builder`

    The `patch_builder` function must take at least two arguments:
        1. A pointer to the original function
        2. The patch policy associated with this patch (may be `None`)
    The `patch_builder` may accept additional positional arguments that are passed to
    this function as a tuple via `builder_args`.

    The `patch_builder` function must return a function that matches the argument
    signature of the original function. The returned function must call the original
    function and return the result.

    Not all patches will have policy. Some patch locations are used solely to apply
    proxies or do library analysis, and so no policy exists for those locations.
    Callers can indicate this by passing "" or None for `loc_name`, in which case no
    patch policy will be retrieved.
    """
    original_func = getattr(owner, attr_name)

    # We don't expect this to ever really happen except possibly in test code;
    # particularly in framework tests. Eventually we may just want to reverse patches
    # between test cases rather than have some awkward logic here.
    if patch_manager.is_patched(original_func) and not force:
        return

    from contrast.agent.policy import registry

    loc_name = owner_name if owner_name is not None else get_name(owner)

    patch_policy = (
        registry.get_policy_by_name(f"{loc_name}.{attr_name}") if loc_name else None
    )

    patch = patch_builder(original_func, patch_policy, *(builder_args or ()))

    patch_manager.patch(owner, attr_name, patch)

    func = add_watermark(patch_manager.as_func(getattr(owner, attr_name)))
    if hasattr(func, "__name__") and not isinstance(func, wrapt.FunctionWrapper):
        func.__name__ = ensure_string(attr_name)


def get_loaded_modules(use_for_patching=False):
    """
    Retrieves and filters all loaded modules

    The parameter `use_for_patching` indicates that this function is being
    called to enable patching. In this case the modules are sorted (to
    provide deterministic behavior) and also the modules_to_ignore list is used.

    NOTE: This method gets called multiple times during the course of agent
    initialization. Ideally it would be called only once for PERF optimization,
    but because sys.modules is global to all threads, we can't guarantee its contents
    will be the same and that a race condition won't happen which would add modules
    across different threads.

    :return: dict of name and module as value
    """
    if not use_for_patching:
        # Have to make a copy of sys.modules in order to avoid RuntimeError: dictionary changed size during iteration
        return {k: v for k, v in dict(sys.modules).items() if inspect.ismodule(v)}

    filtered = OrderedDict()
    filtered.update(
        dict(
            (name, module)
            for name, module in sorted(sys.modules.items())
            if inspect.ismodule(module)
            and not module_is_ignored(name)
            and not is_so_module(module)
        )
    )

    return filtered


def is_so_module(module):
    """
    Return True if module is an .so file, such as
    ".../readline.cpython-38-darwin.so"

    :param module: python module object
    :return: bool
    """
    if not hasattr(module, "__file__") or module.__file__ is None:
        return False

    return module.__file__.endswith(".so")


def is_patchable(obj):
    if inspect.ismodule(obj):
        return False
    if inspect.isclass(obj):
        return True

    # cython methods look like unpatchable builtins, but they can be patched normally
    # an example of this is lxml.etree.fromstring
    # for additional info, see https://groups.google.com/forum/#!topic/cython-users/v5dXFOu-DNc
    is_unpatchable_builtin_method = inspect.ismethoddescriptor(
        obj
    ) and not obj.__class__.__name__.startswith("cython")

    return inspect.isroutine(obj) and not is_unpatchable_builtin_method


@fail_quietly("Unable to repatch single module")
def repatch_module(module):
    """Repatch a single module. See docstring for repatch_imported_modules"""

    module_attrs = list(vars(module).items())

    for attr_name, attr in module_attrs:
        try:
            if not is_patchable(attr):
                continue
        except Exception as e:
            if (
                module.__name__.startswith("django")
                and type(e).__name__ == "ImproperlyConfigured"
            ):
                # Django gives ImproperlyConfigured if present in the env but unused.
                # This is not an issue, but can lead to noisy logging.
                continue
            logger.debug(
                "exception occurred while checking whether to patch %s in %s",
                attr_name,
                module.__name__,
                exc_info=e,
            )
            continue

        if not patch_manager.has_associated_patch(attr):
            continue

        logger.debug("applying repatch to %s in %s", attr_name, module.__name__)
        patch_manager.patch(module, attr_name)


@fail_quietly("Unable to patch previously imported modules")
def repatch_imported_modules():
    """
    Applies patches to modules that were already imported prior to agent startup

    Here's the problem: our patches don't get applied until after our
    middleware class is initialized. At this point it's likely that most (or
    all) application modules will have already been imported.

    If we patch the function `foo.bar.baz`, and an application module that was
    loaded prior to our patches imports it as `from foo.bar import baz`, then
    our patch will have no effect within that application module. This is
    because the application module has a reference to the *original* function,
    and that reference remains unchanged even after we apply a patch to the
    `foo.bar` module.

    On the other hand, if the application imports it as `from foo import bar`
    and uses it as `bar.baz()`, then our patches will work just fine. In this
    case, the application module has a reference to the *module itself*, which
    is where we apply our patch. This means that when the application calls
    `bar.baz()`, it will be calling the updated (patched) function.

    Incidentally, if the application imports as `from foo.bar import baz`, but
    this module is not loaded until *after* our patches have been applied, our
    patch will be effective. However, we have no control over the order of
    imports in an application.

    This function is designed to remedy the former case in order to make sure
    that our patches are effective regardless of how they are imported or the
    order in which they are imported by the application.

    Prior to calling this function, we make a record of every function that
    gets patched. After all patches are applied, this function iterates
    through all imported modules, which includes all modules that may have been
    imported before our patches were applied. We look for any instances of the
    original functions that need to be patched, and we replace them with the
    patches in those modules.
    """
    for _, module in get_loaded_modules(use_for_patching=True).items():
        repatch_module(module)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import requests

from .base_ts_message import BaseTsAppMessage
from contrast.utils.decorators import fail_loudly
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class ApplicationInventory(BaseTsAppMessage):
    def __init__(self, routes):
        # This message does not need "Application-Path" header but it doesn't hurt
        # either.
        super().__init__()
        self.base_url = f"{self.settings.api_url}/agents/v1.0/"

        self.body = {
            "routes": [route.to_json_inventory() for route in routes.values()],
        }

        session_id = self.settings.config.session_id
        if session_id:
            self.body.update({"session_id": session_id})

    @property
    def name(self):
        return "application-inventory"

    @property
    def path(self):
        return "/".join(
            [
                "applications",
                self.server_name_b64,
                self.server_path_b64,
                self.server_type_b64,
                self.app_language_b64,
                self.app_name_b64,
                "inventory",
            ]
        )

    @property
    def request_method(self):
        return requests.post

    @fail_loudly("Failed to process ApplicationInventory response")
    def process_response(self, response, reporting_client):
        self.process_response_code(response, reporting_client)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from typing import Callable, ContextManager

def contrast_scope() -> ContextManager[None]: ...
def propagation_scope() -> ContextManager[None]: ...
def trigger_scope() -> ContextManager[None]: ...
def with_contrast_scope(function: Callable) -> Callable: ...
def in_contrast_scope() -> bool: ...
def in_propagation_scope() -> bool: ...
def in_trigger_scope() -> bool: ...
def enter_contrast_scope() -> None: ...
def exit_contrast_scope() -> None: ...
def enter_propagation_scope() -> None: ...
def exit_propagation_scope() -> None: ...
def enter_trigger_scope() -> None: ...
def exit_trigger_scope() -> None: ...



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from copy import copy

from contrast.agent.assess.tag import Tag
from contrast.utils.assess.tag_utils import (
    combine,
    is_covered,
    ordered_merge,
    intersection,
)


class DefaultAction(object):
    """
    Default action used to evaluate whether a trigger rule is violated

    Most trigger nodes simply need to determine whether any of the trigger node sources
    contain an UNTRUSTED range that isn't also covered by a mitigating tag. However,
    some nodes (e.g. some SSRF triggers) may require more sophisticated logic to
    determine whether the rule was violated. These nodes can specify custom actions
    that implement the necessary logic.

    Custom actions must implement (or inherit) the following four attributes:
        * is_violated
        * find_all_tag_ranges
        * find_any_tag_ranges
        * extract_source
        * unwind_source (boolean)
    """

    def is_violated(self, source_properties, required_tags, disallowed_tags, **kwargs):
        if source_properties is None:
            return False

        # The spec dictates that in order to violate a rule, a range must contain
        # ALL required tags
        vulnerable_tag_ranges = self.find_all_tag_ranges(
            source_properties, required_tags
        )

        if not vulnerable_tag_ranges:
            return False

        # The spec dictates that ANY disallowed tag can mitigate a rule violation
        secure_tag_ranges = self.find_any_tag_ranges(source_properties, disallowed_tags)

        if not secure_tag_ranges:
            return True

        # figure out if there are any vulnerable ranges that aren't
        # covered by a secure one. if there are, the rule was violated
        return not is_covered(vulnerable_tag_ranges, secure_tag_ranges)

    def find_all_tag_ranges(self, properties, desired_tags):
        """
        Find tag ranges that are common to ALL of the given tags

        :param properties:
        :param desired_tags:
        :return: list of Tag instances
        """

        if not properties.is_tracked() or not desired_tags:
            return []

        trimmed = False
        tag_ranges = [Tag(properties.last_tagged_index() + 1, 0)]

        for tag in desired_tags:
            tags = properties.fetch_tags(tag)
            tag_ranges = intersection(tags, tag_ranges)
            trimmed = True

        if not trimmed:
            return []

        return combine(tag_ranges)

    def find_any_tag_ranges(self, properties, desired_tags):
        """
        Find tag ranges that belong to ANY of the given tags

        :param properties:
        :param desired_tags:
        :return: list of Tag instances
        """
        tag_ranges = []

        if not properties.is_tracked() or not desired_tags:
            return tag_ranges

        for tag in desired_tags:
            found = properties.fetch_tags(tag)
            if found:
                tag_ranges = ordered_merge(tag_ranges, copy(found))

        return tag_ranges

    def extract_source(self, source, args, kwargs):
        """
        For some rules (eg ssrf) the source is an object containing a tracked string.
        By default, the source is unmodified.
        This method should be overridden as required.
        """
        return source

    # for sources that are more complex data structures, like lists, dicts, etc., we
    # call `cs__apply_source` once per element by default. Set this to False in a
    # subclass to disable this behavior. This will cause `cs__apply_source` (and
    # `extract_source`) to receive the original, unmodified source object.
    unwind_source = True



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.policy.trigger_actions.default_action import DefaultAction


class OpenAIAction(DefaultAction):
    """
    This is currently no different from the DefaultAction, but that might not always be
    the case. This is ready to go if necessary in the future.
    """

    # for now, keep this True - might change if we modify `extract_source` logic
    unwind_source = True

    def extract_source(  # pylint: disable=useless-parent-delegation
        self, source, args, kwargs
    ):
        """
        The openai API requires a "role" for each prompt message. Currently, we've
        determined that all role types are vulnerable to prompt-injection. If this
        changes in the future, more complicated logic here may be necessary.
        """
        return super().extract_source(source, args, kwargs)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast_vendor import wrapt
from contrast.patches.databases import dbapi2
from contrast.utils import Namespace

ENGINE_DEFAULT_MODULE = "sqlalchemy.engine.default"


class module(Namespace):
    # saved version of the original (unpatched) method
    # this value is None iff the method is currently unpatched
    unpatched_method = None


class CustomFunctionWrapper(wrapt.FunctionWrapper):
    """
    We've seen this before. In python2, instance method objects save their real
    underlying function under their `__func__` attribute. To line up properly with
    SQLAlchemy's instance check, we need type(wrapped_method) to return `function`.

    This is unfortunate because it's probably incorrect, but it's SQLAlchemy's fault.
    Wrapt assumes that accessors aren't purposely bypassing the descriptor protocol
    when accessing a wrapped instance method - but if they do, the behavior is wrong.

    We have to cater to SQLAlchemy here, since it's really unlikely anyone else will
    be type-checking this method.
    """

    @property
    def __class__(self):
        func = getattr(self.__wrapped__, "__func__", self.__wrapped__)
        return func.__class__

    @__class__.setter
    def __class__(self, value):
        self.__wrapped__.__class__ = value


def dialect_init_wrapper(wrapped, instance, args, kwargs):
    """
    Wrapt decides to separate `instance` from `args`. Don't worry that `instance`
    isn't passed back in to `wrapped` at call-time! This is already handled internally
    by wrapt.

    See https://wrapt.readthedocs.io/en/latest/wrappers.html#function-wrappers.
    """
    if "dbapi" in kwargs:
        dbapi2.instrument_adapter(kwargs["dbapi"], "sqlalchemy")
    return wrapped(*args, **kwargs)


def patch_sqlalchemy(sqlalchemy_engine_default):
    """
    SQLAlchemy performs some downright demonic introspection on the __init__ methods
    for DefaultDialect and subclasses (see sqlalchemy.util.langhelpers.get_cls_kwargs).
    The only way we can get around this is to use wrapt's magic to create a very well-
    behaved patch for this method. If we tried to use build_and_apply_patch here, or
    even if we created a naive wrapper for this method, SQLAlchemy's introspection
    would lead to failures internal to the library.

    Hopefully one day this will be our one true patching strategy, as it's
    shockingly robust - it successfully imitates the original method even against
    tactics as powerful as argument introspection.

    Additionally, the patch_manager is not equipped to handle this case. This is
    likely due to the fact that the wrapped method is an instance of a callable class,
    not actually a function. If we want to make use of wrapt wrappers in the future,
    we will need to update the patch_manager accordingly. We also can't check for
    __wrapped__, because DefaultDialect is already wrapped by SQLAlchemy!
    """
    if module.unpatched_method is not None:
        return

    orig_method = sqlalchemy_engine_default.DefaultDialect.__init__
    module.unpatched_method = orig_method
    wrapped_method = CustomFunctionWrapper(orig_method, dialect_init_wrapper)

    setattr(sqlalchemy_engine_default.DefaultDialect, "__init__", wrapped_method)


def register_patches():
    wrapt.register_post_import_hook(patch_sqlalchemy, ENGINE_DEFAULT_MODULE)


def reverse_patches():
    """
    Note: this doesn't reverse-patch any dbapi2 adapters that were patched as a
    result of this module's patch.
    """
    sqlalchemy_engine_default = sys.modules.get(ENGINE_DEFAULT_MODULE)
    if None in (sqlalchemy_engine_default, module.unpatched_method):
        return

    setattr(
        sqlalchemy_engine_default.DefaultDialect, "__init__", module.unpatched_method
    )
    module.unpatched_method = None



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import copy
import inspect

NOTIMPLEMENTED_MSG = "This method should be implemented by concrete subclass subclass"


def safe_copy(value):
    """
    Return a safe copy of a value

    :param value: to be copied
    :return: copied value if no exception
    """
    try:
        return copy.copy(value)
    except Exception:
        return value


def get_name(obj):
    return f"{obj.__module__}.{obj.__name__}" if inspect.isclass(obj) else obj.__name__



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import threading

import contrast
from contrast.reporting import teamserver_messages
from contrast.agent import (
    agent_state,
    scope,
    thread_watcher,
)
from contrast.agent.protect import input_analysis
from contrast.agent.request_context import RequestContext
from contrast.agent.assess.rules.response.xss import analyze_xss
from contrast.agent.assess.rules.response.analyze import analyze_response_rules
from contrast.agent.assess.preflight import update_preflight_hashes
from contrast.agent.middlewares.route_coverage.routes_mixin import RoutesMixin
from contrast.api.attack import ProtectResponse
from contrast.utils.decorators import cached_property
from contrast.utils.loggers.logger import (
    setup_basic_agent_logger,
)
from contrast.utils import timer
from contrast.utils.decorators import fail_loudly, fail_quietly

# initialize a basic logger until config is parsed
logger = setup_basic_agent_logger()


class BaseMiddleware(RoutesMixin):
    """
    BaseMiddleware contains all the initial setup for the framework middlewares

    Requirements:

        1. It's callable
        2. It has call_with_agent
        3. It has call_without_agent

    Pre and post filter calls should not block the flow that this class has.

    Pre -> get_response -> post
    """

    # TODO: PYT-2869 this should be removed. This sort of thing is agent_state's job.
    _loaded = False

    # TODO: PYT-2852 Revisit application name detection
    app_name = ""  # This should be overridden by child classes

    DIAGNOSTIC_ENDPOINT = "/save-contrast-security-config"
    DIAGNOSTIC_ALLOWED_SERVER = "localhost"
    DIAGNOSTIC_ALLOWED_IP = "127.0.0.1"

    @scope.with_contrast_scope
    def __init__(self):
        """
        Most agent initialization is now done by the agent state module

        This method calls agent state initialization if it hasn't been done
        already and then loads the settings and reporting client.

        Config scanning still happens here since the behavior is framework-specific.
        """
        # id will be different across processes but also for multiple middlewares
        # within the same process
        self.id = id(self)
        self.settings = None
        self.request_start_time = None
        self.request_path = None

        if BaseMiddleware._loaded:
            logger.warning(
                "Contrast Agent middleware initialized more than once per process."
            )

        agent_state.initialize()

        self.settings = agent_state.get_settings()
        if not self.is_agent_enabled():
            return

        self.reporting_client = agent_state.get_reporting_client()

        if self.settings.is_assess_enabled() and not BaseMiddleware._loaded:
            BaseMiddleware._loaded = True
            self.scan_configs_in_thread()

        BaseMiddleware._loaded = True

    @cached_property
    def name(self):
        raise NotImplementedError("Must implement name")

    def is_agent_enabled(self):
        """
        Agent is considered enabled if the config value for 'enable' is True (or empty,
        defaults to True). Errors during initialization or runtime will set this value
        to False.
        """
        if self.settings is None:
            return False

        if not self.settings.is_agent_config_enabled():
            return False

        return True

    def call_with_agent(self, *args):
        raise NotImplementedError("Must implement call_with_agent")

    def call_without_agent(self, *args):
        """
        The agent does not set context when this function is called so all other patches
        (e.g propagators) that check context shouldn't run.
        """
        logger.debug("Contrast Agent did not analyze request %s", self.request_path)

    def should_analyze_request(self, environ):
        """
        Determine if request should be analyzed based on configured settings.

        While returning different types of objects based on logic is not a good
        pattern, in this case it's an optimization that allows us to create
        the request context obj when we need it.

        :return: False or RequestContext instance
        """
        if not self.is_agent_enabled():
            logger.debug("Will not analyze request: agent disabled.")
            return False

        context = RequestContext(environ)

        path = environ.get("PATH_INFO")

        if self.settings.evaluate_exclusions(context, path):
            logger.debug("Will not analyze request: request meets exclusions.")
            return False

        from contrast.agent.assess import sampling

        if sampling.enabled() and sampling.meets_criteria(context):
            logger.debug("Will not analyze request: request meets sampling.")
            return False

        return context  # equivalent to returning True

    @fail_loudly("Unable to do handle_ensure")
    def handle_ensure(self, context, request):
        """
        Method that should run for all middlewares AFTER every request is made.
        """
        if context is None:
            logger.error("Context not defined in middleware ensure")
            return

        with scope.contrast_scope():
            thread_watcher.ensure_running(self)

            if self.settings.is_assess_enabled():
                # route discovery and storage needs to occur before sending messages
                self.handle_routes(context, request)

                logger.debug("Updating preflight hashes with route info")
                update_preflight_hashes(context)

            logger.debug("Sending final messages for reporting.")

            for msg in self.final_ts_messages(context):
                self.reporting_client.add_message(msg)

        agent_state.set_first_request(False)

    def final_ts_messages(self, context):
        if context is None:
            return []

        messages = [
            teamserver_messages.ApplicationActivity(context),
        ]

        if (
            self.settings.is_assess_enabled()
            and context.findings is not None
            and len(context.findings) > 0
        ):
            messages.append(
                teamserver_messages.Preflight(context.findings, context.request)
            )

        # Per the spec, we do not report an observed route if the route signature or URL is empty.
        if (
            self.settings.is_assess_enabled()
            and context.observed_route.signature
            and context.observed_route.url
        ):
            messages.append(teamserver_messages.ObservedRoute(context.observed_route))

        # ApplicationUpdate message only sent at first request once we have routes
        if agent_state.is_first_request() and self.settings.is_inventory_enabled():
            messages.append(
                teamserver_messages.ApplicationInventory(agent_state.get_routes())
            )

        # Not every request will have a new dependency loaded, so we can skip sending empty messages.
        if context.observed_library_usage is not None:
            messages.append(context.observed_library_usage)

        return messages

    @fail_loudly("Unable to do prefilter protect")
    def prefilter(self):
        """
        Prefilter - AKA input analysis - is performed mostly with agent-lib but partly
        in the agent.

        This is PROTECT only.

        In this method we call on agent-lib to do input analysis, which can result in:
        1. agent-lib finds an attack in which case we block the request
        2. agent-lib returns input analysis to use for later sink / infilter analysis,
            in which case we store it here in the request context.
        """
        if not self.settings.is_protect_enabled():
            return

        input_analysis.get_input_analysis(prefer_worth_watching=True)
        self.agent_prefilter()

    def agent_prefilter(self):
        """
        Prefilter for any rules that do not yet use agent-lib
        """
        with scope.contrast_scope():
            self.prefilter_protect()

    @fail_quietly("Failed to run prefilter protect.")
    def prefilter_protect(self):
        logger.debug("PROTECT: Running Agent prefilter.")

        for rule in self.settings.protect_rules.values():
            if rule.is_prefilter():
                rule.prefilter()

    @fail_loudly("Unable to do postfilter")
    def postfilter(self, context):
        """
        For all postfilter enabled rules.
        """
        if self.settings.is_protect_enabled():
            self.postfilter_protect(context)

        if self.settings.is_assess_enabled():
            with scope.contrast_scope():
                self.response_analysis(context)

    def _process_trigger_handler(self, handler):
        """
        Gather metadata about response handler callback for xss trigger node

        We need to check whether the response handler callback is an instance method or
        not. This affects the way that our policy machinery works, and it also affects
        reporting, so we need to make sure to account for the possibility that handler
        is a method of some class rather than a standalone function.

        This should be called by the `trigger_node` method in child classes.
        """
        module = handler.__module__
        class_name = ""

        if hasattr(handler, "__self__"):
            class_name = handler.__self__.__class__.__name__
            args = (handler.__self__,)
            instance_method = True
        else:
            args = ()
            instance_method = False

        return module, class_name, args, instance_method

    @cached_property
    def trigger_node(self):
        """
        Trigger node property used by assess reflected xss postfilter rule

        This must be overridden by child classes that make use of the reflected
        xss postfilter rule.
        """
        raise NotImplementedError("Children must define trigger_node property")

    @fail_loudly("Unable to do assess response analysis")
    def response_analysis(self, context):
        """
        Run postfilter for any assess rules. Reflected xss rule runs by default.
        May be overridden in child classes.

        If the response content type matches a allowed content type, do not run
        assess xss response analysis. This is because the security team
        considers reflected xss within these content types to be a false positive.
        """
        logger.debug("ASSESS: Running response analysis")

        analyze_xss(context, self.trigger_node)

        if not self.settings.response_scanning_enabled:
            return

        analyze_response_rules(context)

    @fail_quietly("Failed to run postfilter protect.")
    def postfilter_protect(self, context):
        logger.debug("PROTECT: Running Agent postfilter.")

        input_analysis.get_input_analysis(prefer_worth_watching=False)

        for rule in self.settings.protect_rules.values():
            if rule.is_postfilter():
                rule.postfilter()

    @fail_loudly("Unable to do check_for_blocked")
    def check_for_blocked(self, context: RequestContext):
        """
        Checks for BLOCK events in case SecurityException was caught by app code

        This should be called by each middleware after the view is generated
        but before returning the response (it can be before or after
        postfilter).

        If we make it to this call, it implies that either no SecurityException
        occurred, or if one did occur, it was caught by the application. If we
        find a BLOCK here, it necessarily implies that an attack was detected
        in the application, but the application caught our exception. If the
        application hadn't caught our exception, we never would have made it
        this far because the exception would have already bubbled up to the
        middleware exception handler. So this is really our first and our last
        opportunity to check for this particular edge case.
        """
        for attack in context.attacks:
            if attack.response == ProtectResponse.BLOCKED:
                raise contrast.SecurityException(rule_name=attack.rule_id)

    def log_start_request_analysis(self):
        self.request_start_time = timer.now_ms_float()

        logger.debug("Beginning request analysis", request_path=self.request_path)

    @fail_loudly("Unable to do log_end_request_analysis")
    def log_end_request_analysis(self):
        request_end_time = timer.now_ms_float()

        logger.debug(
            "Ending request analysis",
            request_path=self.request_path,
        )

        elapsed_time = request_end_time - self.request_start_time

        logger.debug(
            "elapsed time request analysis ms",
            elapsed_time=elapsed_time,
            request_path=self.request_path,
        )

    def scan_configs_in_thread(self):
        """
        Run config scanning rules for assess in a separate thread.

        Not all frameworks we support will necessarily have config scanning rules.
        These frameworks should not start a thread.
        """
        if hasattr(self, "_scan_configs"):
            logger.debug("Will start a thread to scan config rules.")
            scanner = threading.Thread(target=self._scan_configs)
            scanner.daemon = True
            scanner.start()
        else:
            logger.debug(f"No config scanning rules for {self.__class__.__name__}")



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.policy.registry import register_trigger_rule
from contrast.agent.policy.utils import CompositeNode


DISALLOWED_TAGS = [
    "CUSTOM_ENCODED_TRUST_BOUNDARY_VIOLATION",
    "CUSTOM_ENCODED",
    "CUSTOM_VALIDATED_TRUST_BOUNDARY_VIOLATION",
    "CUSTOM_VALIDATED",
    "LIMITED_CHARS",
]

unsafe_code_execution_triggers = [
    CompositeNode(
        {
            "module": "BUILTIN",
            "policy_patch": False,
        },
        [
            {
                "method_name": ["exec", "eval"],
                # Takes no keyword arguments
                "source": "ARG_0",
            },
            {
                "method_name": "compile",
                "source": "ARG_0,KWARG:source",
            },
            {
                # This method is also instrumented for library analysis
                "method_name": "__import__",
                "source": "ARG_0,KWARG:name",
            },
        ],
    ),
    {
        "module": "importlib",
        "method_name": "__import__",
        "source": "ARG_0,KWARG:name",
        # Not patched by policy since it's also instrumented for library analysis
        "policy_patch": False,
    },
    {
        "module": "importlib",
        "method_name": "import_module",
        "source": "ARG_0,KWARG:name",
        # TODO: PYT-959 when we instrument this for library analysis, it likely can no longer be a policy patch
        "policy_patch": True,
    },
]


register_trigger_rule(
    DataflowRule.from_nodes(
        "unsafe-code-execution",
        unsafe_code_execution_triggers,
        disallowed_tags=DISALLOWED_TAGS,
    )
)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.policy.registry import register_trigger_rule


unvalidated_redirect_triggers = [
    {
        "module": "django.http.response",
        "class_name": "HttpResponseRedirectBase",
        "method_name": "__init__",
        "source": "ARG_0,KWARG:redirect_to",
    },
    {
        # Base class for redirects in pyramid
        "module": "pyramid.httpexceptions",
        "class_name": "_HTTPMove",
        "method_name": "__init__",
        "source": "ARG_0,KWARG:location",
    },
    {
        # Base class for redirects in webob. Not the same as pyramid
        "module": "webob.exc",
        "class_name": "_HTTPMove",
        "method_name": "__init__",
        "source": "ARG_4,KWARG:location",
    },
    {
        "module": ["werkzeug.utils", "quart.utils"],
        "method_name": "redirect",
        "source": "ARG_0,KWARG:location",
    },
    {
        "module": "bottle",
        "method_name": "redirect",
        "source": "ARG_0,KWARG:url",
    },
    {
        "module": "falcon",
        "class_name": [
            "HTTPMovedPermanently",
            "HTTPFound",
            "HTTPSeeOther",
            "HTTPTemporaryRedirect",
            "HTTPPermanentRedirect",
        ],
        "method_name": "__init__",
        "source": "ARG_0,KWARG:location",
    },
    {
        # Used by fastapi for redirect responses
        "module": "starlette.responses",
        "class_name": "RedirectResponse",
        "method_name": "__init__",
        "source": "ARG_0,KWARG:url",
    },
    {
        "module": "aiohttp.web_exceptions",
        "class_name": "_HTTPMove",
        "method_name": "__init__",
        "source": "ARG_0,KWARG:location",
    },
]


register_trigger_rule(
    DataflowRule.from_nodes(
        "unvalidated-redirect",
        unvalidated_redirect_triggers,
        disallowed_tags=["URL_ENCODED"],
    )
)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
class WrongDjangoMiddlewareException(RuntimeError):
    MSG_FMT = (
        "Incorrect Contrast middleware.\n"
        "You are likely using settings.{setting}, which requires Contrast's {type} DjangoMiddleware.\n"
        "Please use '{middleware}' in settings.{setting}."
    )

    """
    Error message if we see that the application _should_ be using the legacy middleware,
    but is actually using the standard middleware.
    """
    USE_LEGACY_MSG = MSG_FMT.format(
        setting="MIDDLEWARE_CLASSES",
        type="legacy",
        middleware="contrast.agent.middlewares.legacy_django_middleware.DjangoMiddleware",
    )

    """
    Error message if we see that the application _should_ be using the standard middleware,
    but is actually using the legacy middleware
    """
    USE_STANDARD_MSG = MSG_FMT.format(
        setting="MIDDLEWARE",
        type="standard",
        middleware="contrast.agent.middlewares.django_middleware.DjangoMiddleware",
    )

    def __init__(self, using_legacy_middleware):
        # since we don't see our middleware in the expected setting, assume that they are using the other setting
        msg = self.USE_STANDARD_MSG if using_legacy_middleware else self.USE_LEGACY_MSG
        RuntimeError.__init__(self, msg)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

"""
FieldStorage objects and their MiniFieldStorage cousins are data containers for form
and multipart request bodies. The constructor for FieldStorage also parses the raw
request body. We deadzone this constructor but also explicitly track attributes on
the newly created object.

FieldStorage objects can hold one of two things:
    1. a list ("value" or "list") of other FieldStorage / MiniFieldStorage objects;
       even simple form data will have at least one top level FieldStorage with a
       list containing MiniFieldStorage objects for each form param
    2. a field, which has a fieldname ("name"), filename, and a file-like object
       ("file") containing the actual field data. Accessing "value" reads and resets
       the file in this case

MiniFieldStorage objects only hold a parameter name and value. They are used for
storing form data that isn't multipart.

getitem and direct attribute reference is the lower-level API for FieldStorage, but
methods like getvalue(), getfirst(), and getlist() also exist. Since we track the
underlying data directly, we don't need to worry about patches for the higher-level
API; however, we still test it.

The cpython source code for FieldStorage and MiniFieldStorage is written in Python
and the API is not terribly complicated. See the source code (cgi.py) for details.
"""

import sys

import contrast
from contrast.agent import scope
from contrast.agent.policy import patch_manager
from contrast.agent.assess.policy.source_policy import cs__apply_source
from contrast_vendor.wrapt import register_post_import_hook
from contrast.utils.assess import stream_utils
from contrast.utils.decorators import fail_quietly
from contrast.utils.patch_utils import build_and_apply_patch, wrap_and_watermark

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


@fail_quietly("failed to track cgi.FieldStorage object")
def _track_fieldstorage(fs_obj, context, node):
    """
    Explicitly mark all relevant FieldStorage attributes as sources.

    :param fs_obj: FieldStorage object, the instance of cgi.FieldStorage to track
    :param context: RequestContext
    :param node: Source node for cgi.FieldStorage.__init__
    """
    for attr_name in ["name", "filename"]:
        _mark_source(fs_obj, attr_name, context, node)

    if (
        hasattr(fs_obj, "file")
        and fs_obj.file is not None
        and not isinstance(fs_obj.file, stream_utils.BaseStreamProxy)
    ):
        if hasattr(fs_obj.file, "cs__source"):
            fs_obj.file.cs__source = True
        else:
            fs_obj.file = stream_utils.ContrastBufferedReaderProxy(fs_obj.file)
        fs_obj.file.cs__source_type = node.type
        fs_obj.file.cs__source_tags = node.tags


@fail_quietly("failed to track cgi.MiniFieldStorage object")
def _track_minifieldstorage(fs_obj, context, node):
    """
    Same as _track_fieldstorage, but for MiniFieldStorage. This class only stores a
    name and value, both strings.
    """
    for attr_name in ["name", "value"]:
        _mark_source(fs_obj, attr_name, context, node)


def _mark_source(fs_obj, attr_name, context, node):
    attr = getattr(fs_obj, attr_name, None)
    if not attr:
        return
    cs__apply_source(
        context,
        node,
        attr,
        fs_obj,
        attr,
        (),
        {},
        # it's not totally clear what type filename / fieldname should be, so we're
        # sticking with PARAMETER
        source_type="PARAMETER",
        source_name=attr_name,
    )


def _build_patch(original_func, patch_policy, tracker_func):
    """
    Builds the patch for cgi.FieldStorage.__init__ and cgi.MiniFieldStorage.__init__.
    The appropriate tracking function is passed in as tracker_func.
    """
    node = patch_policy.source_nodes[0]

    def init_wrapper(wrapped, instance, args, kwargs):
        # this is a deadzone; we explicitly track the resulting (Mini)FieldStorage
        # object after the fact
        with scope.contrast_scope():
            result = wrapped(*args, **kwargs)

        context = contrast.CS__CONTEXT_TRACKER.current()
        if context is None:
            return result

        tracker_func(instance, context, node)
        return result

    return wrap_and_watermark(original_func, init_wrapper)


def build_cgi_fieldstorage_patch(original_func, patch_policy):
    return _build_patch(original_func, patch_policy, _track_fieldstorage)


def build_cgi_minifieldstorage_patch(original_func, patch_policy):
    return _build_patch(original_func, patch_policy, _track_minifieldstorage)


def patch_cgi(cgi_module):
    build_and_apply_patch(
        cgi_module.FieldStorage, "__init__", build_cgi_fieldstorage_patch
    )
    build_and_apply_patch(
        cgi_module.MiniFieldStorage, "__init__", build_cgi_minifieldstorage_patch
    )


def register_patches():
    register_post_import_hook(patch_cgi, "cgi")


def reverse_patches():
    cgi_module = sys.modules.get("cgi")
    if not cgi_module:
        return

    patch_manager.reverse_patches_by_owner(cgi_module.FieldStorage)
    patch_manager.reverse_patches_by_owner(cgi_module.MiniFieldStorage)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from collections import namedtuple
from html.parser import HTMLParser

from contrast.agent.settings import Settings
from contrast_vendor import structlog as logging
from contrast.utils.string_utils import ensure_string

logger = logging.getLogger("contrast")

Tag = namedtuple("Tag", ["type", "tag", "attrs"])


def analyze_response_rules(context):
    settings = Settings()

    response_rules = settings.enabled_response_rules()
    response = context.response
    if response is None or response.body is None:
        return

    body = ensure_string(response.body)

    if not response_rules or not body:
        return

    content_type = response.headers.get("content-type", "")

    status_code = response.status_code
    valid_response_rules = [
        rule for rule in response_rules if rule.is_valid(status_code, content_type)
    ]

    if not valid_response_rules:
        return

    form_tags, meta_tags = get_tags(body)

    for rule in valid_response_rules:
        violated, properties = rule.is_violated(
            response.headers, body, form_tags, meta_tags
        )
        if violated:
            rule.build_and_append_finding(properties, context)


def get_tags(body):
    form_tags, meta_tags = [], []

    class BodyParser(HTMLParser):
        def handle_starttag(self, tag, attrs):
            if tag == "form":
                form_tags.append(
                    Tag(type="form", tag=self.get_starttag_text(), attrs=attrs)
                )
            if tag == "meta":
                meta_tags.append(
                    Tag(type="meta", tag=self.get_starttag_text(), attrs=attrs)
                )

    parser = BodyParser()
    parser.feed(body)

    return form_tags, meta_tags



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.triggers.httponly_rule import HttpOnlyRule
from contrast.agent.policy.registry import register_trigger_rule


httponly_triggers = [
    {
        "module": "pyramid.session",
        "method_name": "BaseCookieSessionFactory",
        "source": "KWARG:httponly",
        "policy_patch": False,
        "unsafe_default": True,
    },
    {
        # WSGI pluggable for sessions",
        "module": "beaker.session",
        "class_name": "Session",
        "method_name": "__init__",
        "source": "KWARG:httponly",
        "unsafe_default": True,
    },
    {
        # Used by bottle for sessions
        "module": "bottle_session",
        "class_name": "Session",
        "method_name": "__init__",
        "source": "ARG_4,KWARG:cookie_httponly",
        "unsafe_default": True,
    },
    {
        # Used by fastapi for cookies
        "module": "starlette.responses",
        "class_name": "Response",
        "method_name": "set_cookie",
        "source": "ARG_8,KWARG:httponly",
        "unsafe_default": True,
    },
    {
        # Sets custom response cookie and set its secure attribute
        "module": "falcon",
        "class_name": "Response",
        "method_name": "set_cookie",
        "source": "KWARG:http_only",
    },
    {
        # Sets custom response cookie and set its secure attribute
        "module": "aiohttp.web",
        "class_name": "Response",
        "method_name": "set_cookie",
        "source": "KWARG:httponly",
    },
    {
        "module": "starlette_authlib.middleware",
        "class_name": "AuthlibMiddleware",
        "method_name": "__init__",
        "source": "KWARG:https_only",
        "unsafe_default": True,
    },
]


register_trigger_rule(
    HttpOnlyRule.from_nodes(
        "httponly",
        httponly_triggers,
    )
)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import threading
from contrast.agent import scope
from contrast.agent.settings import Settings
from contrast.reporting.teamserver_messages import ServerActivity
from contrast.reporting import ReportingClient
from contrast.utils.decorators import fail_loudly
from contrast.utils.timer import sleep
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")
SERVER_SETTING_THREAD_NAME = "ContrastServerSettings"


class ServerSettingsPoll(threading.Thread):
    def __init__(self):
        self.stopped = False
        self.settings_interval_ms = Settings().config.server_settings_poll_interval
        # Agent should not ping too frequently
        if self.settings_interval_ms < 10000:
            self.settings_interval_ms = 10000

        self.reporting_client = ReportingClient()

        super().__init__()
        # A thread must have had __init__ called, but not start, to set daemon
        self.daemon = True
        self.name = SERVER_SETTING_THREAD_NAME

    def start(self):
        self.stopped = False
        super().start()

    @property
    def settings_interval_sec(self):
        return self.settings_interval_ms / 1000

    def run(self):
        # Ensure the server settings thread runs in scope because it is
        # initialized before our thread.start patch is applied.
        with scope.contrast_scope():
            logger.debug("Establishing Server Settings Thread")

            while not self.stopped and Settings().is_agent_config_enabled():
                self.send_server_settings()
                sleep(self.settings_interval_sec)

    @fail_loudly("Error sending a server settings message")
    def send_server_settings(self):
        """
        We're currently hitting the Server Activity endpoint to get our server
        features/ settings. This NG style endpoint will eventually be replaced
        with the true v1 Server Settings endpoint, so naming around it reflects
        that. The Server Activity message is always empty, so we're already
        only using it for settings, not reporting. The content of the message
        is for now deprecated java-only features.
        """
        if Settings().config is None:
            return
        msg = ServerActivity()
        response = self.reporting_client.send_message(msg)
        msg.process_response(response, self.reporting_client)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.policy.registry import register_trigger_rule


DISALLOWED_TAGS = [
    "CUSTOM_ENCODED_XPATH_INJECTION",
    "CUSTOM_ENCODED",
    "CUSTOM_VALIDATED_XPATH_INJECTION",
    "CUSTOM_VALIDATED",
    "LIMITED_CHARS",
    "XPATH_ENCODED",
]


xpath_injection_triggers = [
    {
        "module": "xml.etree.ElementTree",
        "class_name": "Element",
        "method_name": ["find", "findall", "findtext"],
        "source": "ARG_0",
    },
    {
        "module": "lxml.etree",
        "class_name": "_Element",
        "method_name": ["find", "findall", "findtext", "xpath"],
        "source": "ARG_0",
    },
    {
        "module": "lxml.etree",
        "class_name": "_ElementTree",
        "method_name": "xpath",
        "source": "ARG_0",
    },
    {
        "module": "lxml.etree",
        "class_name": "XPath",
        "method_name": "__init__",
        "source": "ARG_0",
        "policy_patch": False,
    },
    {
        "module": "lxml.etree",
        "class_name": "XPathElementEvaluator",
        "method_name": "__call__",
        "source": "ARG_0",
        "policy_patch": False,
    },
    {
        "module": "lxml.etree",
        "class_name": "XPathDocumentEvaluator",
        "method_name": "__call__",
        "source": "ARG_0",
        "policy_patch": False,
    },
]


register_trigger_rule(
    DataflowRule.from_nodes(
        "xpath-injection",
        xpath_injection_triggers,
        disallowed_tags=DISALLOWED_TAGS,
    )
)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_propagation_nodes
from contrast.agent.policy.utils import CompositeNode


encodings_propagators = [
    {
        "module": [
            f"encodings.{name}"
            for name in [
                "ascii",
                "latin_1",
                "unicode_escape",
                "raw_unicode_escape",
                "unicode_internal",
                "utf8",
            ]
        ],
        "class_name": "Codec",
        # None of these methods have keyword arguments
        "method_name": ["encode", "decode"],
        "instance_method": False,
        "policy_patch": False,
        "source": "ARG_0",
        "target": "RETURN",
        "action": "CODECS_SPLAT",
    },
    CompositeNode(
        {
            "module": "binascii",
            "target": "RETURN",
            "action": "SPLAT",
        },
        [
            {
                "method_name": "a2b_base64",
                "untags": ["BASE64_ENCODED"],
                # no keyword arguments
                "source": "ARG_0",
            },
            {
                "method_name": "b2a_base64",
                "tags": ["BASE64_ENCODED"],
                # no keyword arguments
                "source": "ARG_0",
            },
            {
                "method_name": ["hexlify", "unhexlify", "b2a_hex", "a2b_hex"],
                # no keyword arguments
                "source": "ARG_0",
            },
        ],
    ),
    CompositeNode(
        {
            "module": "base64",
            "target": "RETURN",
            "action": "SPLAT",
        },
        [
            {
                # In Py3.10 this has been renamed to _b32decode
                "method_name": "b32decode",
                "source": "ARG_0,KWARG:s",
                "untags": ["BASE64_ENCODED"],
            },
            {
                # In Py3.10 this has been renamed to _b32encode
                "method_name": "b32encode",
                "source": "ARG_0,KWARG:s",
                "tags": ["BASE64_ENCODED"],
            },
            {
                "method_name": "_b32decode",
                "source": "ARG_1,KWARG:s",
                "untags": ["BASE64_ENCODED"],
            },
            {
                "method_name": "_b32encode",
                "source": "ARG_1,KWARG:s",
                "tags": ["BASE64_ENCODED"],
            },
            {
                "method_name": "a85decode",
                "source": "ARG_0,KWARG:b",
                "untags": ["BASE64_ENCODED"],
            },
            {
                "method_name": "a85encode",
                "source": "ARG_0,KWARG:b",
                "tags": ["BASE64_ENCODED"],
            },
            {
                "method_name": "b85decode",
                "source": "ARG_0,KWARG:b",
                "untags": ["BASE64_ENCODED"],
            },
            {
                "method_name": "b85encode",
                "source": "ARG_0,KWARG:b",
                "tags": ["BASE64_ENCODED"],
            },
        ],
    ),
]

register_propagation_nodes(encodings_propagators)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast_vendor.wrapt import register_post_import_hook

from contrast.patches.databases import dbapi2

PYMYSQL = "pymysql"
VENDOR = "MySQL"


def instrument_pymysql(pymysql):
    dbapi2.instrument_adapter(
        pymysql, VENDOR, dbapi2.Dbapi2Patcher, extra_cursors=[pymysql.cursors.Cursor]
    )


def register_patches():
    register_post_import_hook(instrument_pymysql, PYMYSQL)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys
import functools
from contrast_vendor.wrapt import register_post_import_hook

import contrast
from contrast.agent import scope
from contrast.agent.middlewares.route_coverage.common import build_route
from contrast.agent.middlewares.route_coverage.flask_routes import create_routes
from contrast.agent.policy import patch_manager
from contrast.utils.patch_utils import build_and_apply_patch, wrap_and_watermark
from contrast.utils.decorators import fail_quietly
from contrast.utils.safe_import import safe_import_list

from contrast.agent.assess.rules.config import (
    FlaskSessionAgeRule,
    FlaskSecureFlagRule,
    FlaskHttpOnlyRule,
)

from contrast_vendor import structlog as logging

FLASK_MODULE_NAME = "flask"
QUART_MODULE_NAME = "quart"

logger = logging.getLogger("contrast")


@functools.lru_cache(maxsize=1)
def _get_flask_app_type():
    return tuple(safe_import_list("flask.Flask"))


@functools.lru_cache(maxsize=1)
def _get_quart_app_type():
    return tuple(safe_import_list("quart.Quart"))


def build_flask_full_dispatch_request_patch(orig_func, patch_policy):
    del patch_policy

    def full_dispatch_request_patch(wrapped, instance, args, kwargs):
        try:
            result = wrapped(*args, **kwargs)
        finally:
            do_first_request_analysis(instance)
            do_flask_route_observation(instance)
        return result

    return wrap_and_watermark(orig_func, full_dispatch_request_patch)


def build_quart_full_dispatch_request_patch(orig_func, patch_policy):
    del patch_policy

    async def full_dispatch_request_patch(wrapped, instance, args, kwargs):
        try:
            result = await wrapped(*args, **kwargs)
        finally:
            do_first_request_analysis(instance)
            do_quart_route_observation(instance, *args, **kwargs)
        return result

    return wrap_and_watermark(orig_func, full_dispatch_request_patch)


@fail_quietly("Failed to run first-request analysis")
@scope.with_contrast_scope
def do_first_request_analysis(app_instance):
    from contrast.agent import agent_state

    if not agent_state.is_first_request():
        return

    do_config_scanning(app_instance)
    do_route_discovery(app_instance)


@fail_quietly("unable to perform Flask route observation")
@scope.with_contrast_scope
def do_flask_route_observation(flask_instance):
    logger.debug("Performing Flask route observation")

    flask_ctx = None
    try:
        from flask.globals import request_ctx

        flask_ctx = request_ctx
    except ImportError:
        from flask.globals import _request_ctx_stack

        flask_ctx = _request_ctx_stack.top

    do_route_observation(flask_ctx, flask_instance)


@fail_quietly("unable to perform Quart route observation")
@scope.with_contrast_scope
def do_quart_route_observation(quart_instance, *args, **kwargs):
    logger.debug("Performing Quart route observation")

    quart_ctx = args[0] if len(args) > 0 else kwargs.get("request_context")
    do_route_observation(quart_ctx, quart_instance)


@fail_quietly("unable to perform Flask/Quart route observation")
@scope.with_contrast_scope
def do_route_observation(framework_ctx, app_instance):
    if not framework_ctx:
        logger.debug("unable to get framework ctx for route observation")
        return

    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None:
        logger.debug("not in request context - skipping route observation")
        return

    endpoint = getattr(framework_ctx.request.url_rule, "endpoint", None)
    view_func = app_instance.view_functions.get(endpoint)
    if view_func is None:
        logger.debug("did not find endpoint for route observation")
        return

    context.view_func = view_func
    context.view_func_str = build_route(view_func.__name__, view_func)
    logger.debug("Observed route: %s", context.view_func_str)


@fail_quietly("Failed to run config scanning rules")
def do_config_scanning(app_instance):
    logger.debug("Running config scanning rules")
    for rule in [FlaskSessionAgeRule, FlaskSecureFlagRule, FlaskHttpOnlyRule]:
        rule().apply(app_instance)


@fail_quietly("unable to perform route discovery")
def do_route_discovery(app_instance):
    from contrast.agent import agent_state

    discovered_routes = create_routes(app_instance)
    agent_state.update_routes(discovered_routes)
    logger.debug(
        "Discovered the following routes: %s",
        [f"{route.verb} {route.url}" for route in discovered_routes.values()],
    )


def patch_flask(flask_module):
    build_and_apply_patch(
        flask_module.Flask,
        "full_dispatch_request",
        build_flask_full_dispatch_request_patch,
    )


def patch_quart(quart_module):
    build_and_apply_patch(
        quart_module.Quart,
        "full_dispatch_request",
        build_quart_full_dispatch_request_patch,
    )


def register_patches():
    register_post_import_hook(patch_flask, FLASK_MODULE_NAME)
    register_post_import_hook(patch_quart, QUART_MODULE_NAME)


def reverse_patches():
    flask_module = sys.modules.get(FLASK_MODULE_NAME)
    if flask_module:
        patch_manager.reverse_patches_by_owner(flask_module.Flask)

    quart_module = sys.modules.get(QUART_MODULE_NAME)
    if quart_module:
        patch_manager.reverse_patches_by_owner(quart_module.Quart)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast.agent.assess.rules.config.base_config_rule import BaseConfigRule
from contrast.utils.decorators import fail_quietly


class BaseFlaskConfigRule(BaseConfigRule):
    # This should be overridden by child classes
    SETTINGS_VALUE = ""

    @property
    def name(self):
        raise NotImplementedError

    def get_config_value(self, config):
        return config.get(self.SETTINGS_VALUE)

    def is_violated(self, value):
        raise NotImplementedError

    def get_path(self, app):
        app_module = sys.modules.get(app.name)
        if not app_module:
            return ""

        return app_module.__file__

    def get_snippet(self, value):
        return f'app.config["{self.SETTINGS_VALUE}"] = {value!r}'

    @fail_quietly("Failed to apply config rule")
    def apply(self, app):
        value = self.get_config_value(app.config)
        if not self.is_violated(value):
            return

        config_path = self.get_path(app)

        properties = self.create_properties(value, config_path)

        self.build_and_send_finding(
            properties, settings_module_path=(config_path or app.name)
        )

    def update_preflight_hash(self, hasher, settings_module_path=""):
        """
        Override method in base class for custom preflight hash generation
        """
        hasher.update(settings_module_path)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

from contrast.agent.settings import Settings
from contrast.utils.decorators import fail_quietly
from contrast.utils.timer import now_ms

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")

# Safe to look for settings at module level because sampling feature
# should always run during an agent-backed request.
settings = Settings()

REQUESTS = dict()
LOGGED_SAMPLING_MESSAGE = False


def enabled():
    global LOGGED_SAMPLING_MESSAGE
    enabled = settings.is_assess_enabled() and settings.config.get(
        "assess.sampling.enable"
    )

    if not LOGGED_SAMPLING_MESSAGE and enabled:
        logger.info(
            "Contrast assess.sampling is enabled in your configuration. "
            "Not all requests will be analyzed."
        )
        LOGGED_SAMPLING_MESSAGE = True

    return enabled


# If we can't determine sampling, it's better to analyze the request.
@fail_quietly("Unable to determine sampling", False)
def meets_criteria(context) -> bool:
    """
    If a request meets criteria for sampling, agent should not analyze request.

    Criteria is met if request happens inside a sampling time window and request
    count minus baseline doesn't meet frequency setting.
    """
    if context.hash is None:
        # If we cannot determine the uniqueness of this request, it's better
        # to analyze the request
        return False

    if context.hash in REQUESTS:
        history = REQUESTS[context.hash]
    else:
        history = REQUESTS[context.hash] = RequestHistory()

    history.hit()

    # if sampling window has been exceeded, reset it.
    if history.elapsed() >= settings.config.get("assess.sampling.window_ms"):
        history.reset()
        return False

    # once hits exceed baseline setting, limit analysis based on frequency setting.
    baseline = settings.config.get("assess.sampling.baseline")
    frequency = settings.config.get("assess.sampling.request_frequency")

    return not (history.hits <= baseline or (history.hits - baseline) % frequency == 0)


class RequestHistory(object):
    def __init__(self):
        self._start = now_ms()
        self._hit = 0

    @property
    def hits(self):
        return self._hit

    def elapsed(self):
        return now_ms() - self._start

    def hit(self):
        """Increment attempted requests, not analyzed requests"""
        self._hit += 1

    def reset(self):
        self._start = now_ms()
        self._hit = 1

    def __repr__(self):
        return f"Started at: {self._start} - Hit : {self._hit}"



# SPDX-License-Identifier: MIT OR Apache-2.0
# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the MIT License.  See the LICENSE file in the root of this
# repository for complete details.

"""
Processors and tools specific to the `Twisted <https://twisted.org/>`_
networking engine.

See also :doc:`structlog's Twisted support <twisted>`.
"""

from __future__ import annotations

import json
import sys

from typing import Any, Callable, Sequence, TextIO

from twisted.python import log
from twisted.python.failure import Failure
from twisted.python.log import ILogObserver, textFromEventDict
from zope.interface import implementer

from ._base import BoundLoggerBase
from ._config import _BUILTIN_DEFAULT_PROCESSORS
from ._utils import until_not_interrupted
from .processors import JSONRenderer as GenericJSONRenderer
from .typing import EventDict, WrappedLogger


class BoundLogger(BoundLoggerBase):
    """
    Twisted-specific version of `structlog.BoundLogger`.

    Works exactly like the generic one except that it takes advantage of
    knowing the logging methods in advance.

    Use it like::

        configure(
            wrapper_class=structlog.twisted.BoundLogger,
        )

    """

    def msg(self, event: str | None = None, **kw: Any) -> Any:
        """
        Process event and call ``log.msg()`` with the result.
        """
        return self._proxy_to_logger("msg", event, **kw)

    def err(self, event: str | None = None, **kw: Any) -> Any:
        """
        Process event and call ``log.err()`` with the result.
        """
        return self._proxy_to_logger("err", event, **kw)


class LoggerFactory:
    """
    Build a Twisted logger when an *instance* is called.

    >>> from contrast_vendor.structlog import configure
    >>> from contrast_vendor.structlog.twisted import LoggerFactory
    >>> configure(logger_factory=LoggerFactory())
    """

    def __call__(self, *args: Any) -> WrappedLogger:
        """
        Positional arguments are silently ignored.

        :rvalue: A new Twisted logger.

        .. versionchanged:: 0.4.0
            Added support for optional positional arguments.
        """
        return log


_FAIL_TYPES = (BaseException, Failure)


def _extractStuffAndWhy(eventDict: EventDict) -> tuple[Any, Any, EventDict]:
    """
    Removes all possible *_why*s and *_stuff*s, analyzes exc_info and returns
    a tuple of ``(_stuff, _why, eventDict)``.

    **Modifies** *eventDict*!
    """
    _stuff = eventDict.pop("_stuff", None)
    _why = eventDict.pop("_why", None)
    event = eventDict.pop("event", None)

    if isinstance(_stuff, _FAIL_TYPES) and isinstance(event, _FAIL_TYPES):
        raise ValueError("Both _stuff and event contain an Exception/Failure.")

    # `log.err('event', _why='alsoEvent')` is ambiguous.
    if _why and isinstance(event, str):
        raise ValueError("Both `_why` and `event` supplied.")

    # Two failures are ambiguous too.
    if not isinstance(_stuff, _FAIL_TYPES) and isinstance(event, _FAIL_TYPES):
        _why = _why or "error"
        _stuff = event

    if isinstance(event, str):
        _why = event

    if not _stuff and sys.exc_info() != (None, None, None):
        _stuff = Failure()  # type: ignore[no-untyped-call]

    # Either we used the error ourselves or the user supplied one for
    # formatting.  Avoid log.err() to dump another traceback into the log.
    if isinstance(_stuff, BaseException) and not isinstance(_stuff, Failure):
        _stuff = Failure(_stuff)  # type: ignore[no-untyped-call]

    return _stuff, _why, eventDict


class ReprWrapper:
    """
    Wrap a string and return it as the ``__repr__``.

    This is needed for ``twisted.python.log.err`` that calls `repr` on
    ``_stuff``:

    >>> repr("foo")
    "'foo'"
    >>> repr(ReprWrapper("foo"))
    'foo'

    Note the extra quotes in the unwrapped example.
    """

    def __init__(self, string: str) -> None:
        self.string = string

    def __eq__(self, other: Any) -> bool:
        """
        Check for equality, just for tests.
        """
        return (
            isinstance(other, self.__class__) and self.string == other.string
        )

    def __repr__(self) -> str:
        return self.string


class JSONRenderer(GenericJSONRenderer):
    """
    Behaves like `structlog.processors.JSONRenderer` except that it formats
    tracebacks and failures itself if called with ``err()``.

    .. note::

        This ultimately means that the messages get logged out using ``msg()``,
        and *not* ``err()`` which renders failures in separate lines.

        Therefore it will break your tests that contain assertions using
        `flushLoggedErrors
        <https://docs.twisted.org/en/stable/api/
        twisted.trial.unittest.SynchronousTestCase.html#flushLoggedErrors>`_.

    *Not* an adapter like `EventAdapter` but a real formatter.  Also does *not*
    require to be adapted using it.

    Use together with a `JSONLogObserverWrapper`-wrapped Twisted logger like
    `plainJSONStdOutLogger` for pure-JSON logs.
    """

    def __call__(  # type: ignore[override]
        self,
        logger: WrappedLogger,
        name: str,
        eventDict: EventDict,
    ) -> tuple[Sequence[Any], dict[str, Any]]:
        _stuff, _why, eventDict = _extractStuffAndWhy(eventDict)
        if name == "err":
            eventDict["event"] = _why
            if isinstance(_stuff, Failure):
                eventDict["exception"] = _stuff.getTraceback(detail="verbose")
                _stuff.cleanFailure()  # type: ignore[no-untyped-call]
        else:
            eventDict["event"] = _why
        return (
            (
                ReprWrapper(
                    GenericJSONRenderer.__call__(  # type: ignore[arg-type]
                        self, logger, name, eventDict
                    )
                ),
            ),
            {"_structlog": True},
        )


@implementer(ILogObserver)
class PlainFileLogObserver:
    """
    Write only the the plain message without timestamps or anything else.

    Great to just print JSON to stdout where you catch it with something like
    runit.

    :param file: File to print to.

    .. versionadded:: 0.2.0
    """

    def __init__(self, file: TextIO) -> None:
        self._write = file.write
        self._flush = file.flush

    def __call__(self, eventDict: EventDict) -> None:
        until_not_interrupted(
            self._write,
            textFromEventDict(eventDict)  # type: ignore[arg-type, operator]
            + "\n",
        )
        until_not_interrupted(self._flush)


@implementer(ILogObserver)
class JSONLogObserverWrapper:
    """
    Wrap a log *observer* and render non-`JSONRenderer` entries to JSON.

    :param ILogObserver observer: Twisted log observer to wrap.  For example
        :class:`PlainFileObserver` or Twisted's stock `FileLogObserver
        <https://docs.twisted.org/en/stable/api/
        twisted.python.log.FileLogObserver.html>`_

    .. versionadded:: 0.2.0
    """

    def __init__(self, observer: Any) -> None:
        self._observer = observer

    def __call__(self, eventDict: EventDict) -> str:
        if "_structlog" not in eventDict:
            eventDict["message"] = (
                json.dumps(
                    {
                        "event": textFromEventDict(
                            eventDict  # type: ignore[arg-type]
                        ),
                        "system": eventDict.get("system"),
                    }
                ),
            )
            eventDict["_structlog"] = True

        return self._observer(eventDict)


def plainJSONStdOutLogger() -> JSONLogObserverWrapper:
    """
    Return a logger that writes only the message to stdout.

    Transforms non-`JSONRenderer` messages to JSON.

    Ideal for JSONifying log entries from Twisted plugins and libraries that
    are outside of your control::

        $ twistd -n --logger structlog.twisted.plainJSONStdOutLogger web
        {"event": "Log opened.", "system": "-"}
        {"event": "twistd 13.1.0 (python 2.7.3) starting up.", "system": "-"}
        {"event": "reactor class: twisted...EPollReactor.", "system": "-"}
        {"event": "Site starting on 8080", "system": "-"}
        {"event": "Starting factory <twisted.web.server.Site ...>", ...}
        ...

    Composes `PlainFileLogObserver` and `JSONLogObserverWrapper` to a usable
    logger.

    .. versionadded:: 0.2.0
    """
    return JSONLogObserverWrapper(PlainFileLogObserver(sys.stdout))


class EventAdapter:
    """
    Adapt an ``event_dict`` to Twisted logging system.

    Particularly, make a wrapped `twisted.python.log.err
    <https://docs.twisted.org/en/stable/api/twisted.python.log.html#err>`_
    behave as expected.

    :param dictRenderer: Renderer that is used for the actual log message.
        Please note that structlog comes with a dedicated `JSONRenderer`.

    **Must** be the last processor in the chain and requires a *dictRenderer*
    for the actual formatting as an constructor argument in order to be able to
    fully support the original behaviors of ``log.msg()`` and ``log.err()``.
    """

    def __init__(
        self,
        dictRenderer: Callable[[WrappedLogger, str, EventDict], str]
        | None = None,
    ) -> None:
        """
        :param dictRenderer: A processor used to format the log message.
        """
        self._dictRenderer = dictRenderer or _BUILTIN_DEFAULT_PROCESSORS[-1]

    def __call__(
        self, logger: WrappedLogger, name: str, eventDict: EventDict
    ) -> Any:
        if name == "err":
            # This aspires to handle the following cases correctly:
            #   - log.err(failure, _why='event', **kw)
            #   - log.err('event', **kw)
            #   - log.err(_stuff=failure, _why='event', **kw)
            _stuff, _why, eventDict = _extractStuffAndWhy(eventDict)
            eventDict["event"] = _why

            return (
                (),
                {
                    "_stuff": _stuff,
                    "_why": self._dictRenderer(logger, name, eventDict),
                },
            )

        return self._dictRenderer(logger, name, eventDict)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.triggers.session_timeout_rule import SessionTimeoutRule
from contrast.agent.policy.registry import register_trigger_rule


session_timeout_triggers = [
    {
        # default value is 1200s which is safe
        "module": "pyramid.session",
        "method_name": "BaseCookieSessionFactory",
        "source": "KWARG:timeout",
        "policy_patch": False,
    },
    {
        # WSGI pluggable for sessions
        "module": "beaker.session",
        "class_name": "Session",
        "method_name": "__init__",
        "source": "KWARG:timeout",
        "unsafe_default": True,
    },
    {
        # Used by bottle for sessions
        "module": "bottle_session",
        "class_name": "Session",
        "method_name": "__init__",
        "source": "ARG_2,KWARG:cookie_lifetime",
        "unsafe_default": True,
    },
    {
        "module": "falcon",
        "class_name": "Response",
        "method_name": "set_cookie",
        "source": "KWARG:max_age,KWARG:expires",
    },
    {
        # Used by fastapi for cookies
        "module": "starlette.responses",
        "class_name": "Response",
        "method_name": "set_cookie",
        "source": "KWARG:expires,KWARG:max_age",
    },
    {
        "module": "starlette.middleware.sessions",
        "class_name": "SessionMiddleware",
        "method_name": "__init__",
        "source": "KWARG:max_age",
        "unsafe_default": True,
    },
    {
        "module": "starlette_authlib.middleware",
        "class_name": "AuthlibMiddleware",
        "method_name": "__init__",
        "source": "KWARG:max_age",
        "unsafe_default": True,
    },
    {
        # Set custom response cookie and set its expires or max_age attribute
        "module": "aiohttp.web",
        "class_name": "Response",
        "method_name": "set_cookie",
        "source": "KWARG:expires,KWARG:max_age",
        "unsafe_default": True,
    },
]


register_trigger_rule(
    SessionTimeoutRule.from_nodes("session-timeout", session_timeout_triggers)
)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_source_nodes


cgi_sources = [
    {
        # We actually track attributes of the resulting FieldStorage object, so the
        # target and type here aren't really accurate. See the explicit patch
        "module": "cgi",
        "class_name": ["FieldStorage", "MiniFieldStorage"],
        "method_name": "__init__",
        "policy_patch": False,
        "target": "OBJ",
        "node_type": "MULTIPART_FORM_DATA",
        "tags": ["CROSS_SITE"],
    },
]


register_source_nodes(cgi_sources)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from inspect import ismethoddescriptor, ismodule

from contrast.agent.assess.rules.static_rule import StaticRule
from contrast.agent.settings import Settings
from contrast.utils.decorators import fail_quietly
from contrast.utils.stdlib_modules import is_stdlib_module
from .code_parser_mixin import CodeParserMixin

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class HardcodedValueRule(StaticRule, CodeParserMixin):
    MODULES_TO_IGNORE = [
        "contrast",
        "django",
        "flask",
        "jinja2",
        "pyramid",
        "webob",
        "werkzeug",
    ]

    @property
    def name(self):
        return "hardcoded-value-rule"

    def applies_to(self, _):
        settings = Settings()

        return settings.is_assess_enabled() and not self.disabled

    def is_name_valid(self, constant):
        raise NotImplementedError

    def is_value_type_valid(self, value):
        """
        A value has a value type if it it a str or bytes
        """
        return isinstance(value, (bytes, str, bytearray))

    def is_value_valid(self, value):
        return True

    @fail_quietly("Unable to analyze module")
    def analyze(self, cls_or_module):
        if self.disabled:
            return

        if self._should_ignore_module(cls_or_module):
            return

        attributes = dir(cls_or_module)

        # keep track of parsed code if it has been parsed already
        # to not parse it multiple times for the same cls_or_module
        code_ast = None
        for attribute in attributes:
            if self.is_private(attribute):
                continue

            if self.is_name_valid(attribute):
                continue

            attr_value = getattr(cls_or_module, attribute)

            if ismethoddescriptor(attr_value):
                continue

            if not self.is_value_type_valid(attr_value):
                continue

            if not self.is_value_valid(attr_value):
                continue

            hardcoded_found, code_ast = self.find_hardcoded_str(
                cls_or_module, attribute, code_ast
            )
            if not hardcoded_found:
                continue

            self.report_finding(cls_or_module, attribute)

    # The name of the field
    CONSTANT_NAME = "name"
    # The code line, recreated, with the password obfuscated
    CODE_SOURCE = "codeSource"
    # The constant name
    SOURCE = "source"

    def update_preflight_hash(self, hasher, source_name=None, constant=None, **kwargs):
        """
        Override method in base class for custom preflight hash generation
        """
        hasher.update(source_name)
        hasher.update(constant)

    def report_finding(self, cls_or_module, constant):
        if ismodule(cls_or_module):
            source = cls_or_module.__file__
            source_name = cls_or_module.__name__
            constant_name = constant
        else:
            source = sys.modules[cls_or_module.__module__].__file__
            source_name = f"{cls_or_module.__module__}.{cls_or_module.__name__}"
            constant_name = f"{cls_or_module.__name__}.{constant}"

        logger.debug("Reporting finding in %s for %s", source_name, constant)

        properties = {}
        properties[self.SOURCE] = source
        properties[self.CONSTANT_NAME] = constant
        properties[self.CODE_SOURCE] = f"{constant_name}{self.redacted_marker}"

        self.build_and_send_finding(
            properties, source_name=source_name, constant=constant
        )

    @property
    def redacted_marker(self):
        return " = [**REDACTED**]"

    def _should_ignore_module(self, cls_or_module):
        """
        Return True if the module (or the class's module) should be ignored.
        """
        if ismodule(cls_or_module):
            name = cls_or_module.__name__
        else:
            name = cls_or_module.__module__

        if is_stdlib_module(name):
            return True

        module_name = name.lower() + "."
        if any(module_name.startswith(item) for item in self.MODULES_TO_IGNORE):
            return True

        return False

    def is_private(self, attribute):
        """
        A class or module attribute is private if
        it's prefixed with __ or _ or  cs__
        :param attribute: str representing an attribute
        :return: bool
        """
        return (
            attribute.startswith("__")
            or attribute.startswith("_")
            or attribute.startswith("cs__")
        )

    def fuzzy_match(self, constant_value, field_names):
        lower_cont = constant_value.lower()
        return [
            lower_cont.startswith(name.lower()) or lower_cont.endswith(name.lower())
            for name in field_names
        ]



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.disable_reaction import DisableReaction

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


class ReactionProcessor(object):
    @staticmethod
    def process(application_settings, settings):
        if isinstance(application_settings, list):
            ts_reactions = application_settings

            for reaction in ts_reactions:
                operation = reaction.get("operation", "")
                msg = reaction.get("message", "")

                logger.debug("Received the following reaction: %s %s", operation, msg)

                if operation == DisableReaction.NAME:
                    DisableReaction.run(settings)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .application_activity import ApplicationActivity
from .application_inventory import ApplicationInventory
from .application_startup import ApplicationStartup
from .application_update import ApplicationUpdate
from .heartbeat import Heartbeat
from .library_usage import LibraryUsage
from .observed_route import ObservedRoute
from .preflight import Preflight
from .server_activity import ServerActivity
from .server_startup import ServerStartup
from .base_ts_message import BaseTsMessage, BaseTsAppMessage, BaseTsServerMessage



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent import scope
from contrast.agent.assess.policy import trigger_policy
from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def analyze_xss(context, trigger_node):
    logger.debug("ASSESS: Running xss response analysis")

    # it is possible other forms (capital case) variants exist
    accepted_xss_response_content_types = [
        "/csv",
        "/javascript",
        "/json",
        "/pdf",
        "/x-javascript",
        "/x-json",
        "/plain",
    ]

    content_type = context.response.headers.get("content-type", "")

    if not any(
        [name for name in accepted_xss_response_content_types if name in content_type]
    ):
        apply_policy(context, trigger_node)


def apply_policy(context, trigger_node):
    """
    Evaluate xss rule for assess

    The xss rule is applied to the response body in order to determine whether it
    contains untrusted data. We rely on propagation through any template rendering
    and through the framework. The expectation is that untrusted data from
    a request will propagate successfully all the way to the response body, which
    we are able to see here.

    Each child middleware class must implement specific logic for the trigger node
    since the reporting will differ between frameworks.
    """
    from contrast.agent.policy import registry

    rule = registry.get_triggers_by_rule("reflected-xss")

    # We need to exit scope here in order to account for the fact that some
    # frameworks evaluate the content lazily. We don't want to be in scope when
    # that occurs since it would make us lose propagation. This would prevent us
    # from seeing the response as a tracked string, which we require in order to
    # apply the rule.
    with scope.pop_contrast_scope():
        result = context.response.body

    trigger_node, args = trigger_node

    trigger_policy.apply(rule, [trigger_node], result, args)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.api import Route
from contrast.agent.middlewares.route_coverage.common import (
    build_route,
    DEFAULT_ROUTE_METHODS,
    get_normalized_uri,
    build_key,
)

DEFAULT_ROUTE_METHODS = DEFAULT_ROUTE_METHODS + ("PUT", "PATCH", "DELETE")


def create_bottle_routes(app):
    """
    Returns all the routes registered to a Bottle app as a dict
    :param app: Bottle app
    :return: dict {route_id:  api.Route}
    """
    routes = {}
    for rule in app.routes:
        view_func = rule.callback
        route = build_route(rule.rule, view_func)
        route_id = str(id(view_func))

        for method_type in DEFAULT_ROUTE_METHODS:
            key = build_key(route_id, method_type)
            routes[key] = Route(
                verb=method_type,
                url=get_normalized_uri(str(rule.rule)),
                route=route,
            )
    return routes



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .middleware import FlaskMiddleware as ContrastMiddleware  # noqa



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from .cs_str import set_attr_on_type
from .cs_str import is_string_interned

SPECIAL_METHOD_EXCEPTIONS = ["__repr__", "__getitem__"]


def smart_setattr(owner, name, patch):
    """
    Use either setattr or set_attr_on_type as appropriate
    """
    # For some reason, set_attr_on_type doesn't work for special methods
    # We may eventually need to refine this logic
    force_patch = isinstance(owner, type) and not (
        name not in SPECIAL_METHOD_EXCEPTIONS
        and name.startswith("__")
        and name.endswith("__")
    )

    setattr(owner, name, patch) if not force_patch else set_attr_on_type(
        owner, name, patch
    )


def is_interned(value):
    """
    Checks if a string is interned,
    uses cpython implemented function is_string_interned()
    Arguments:
        value: a string or byte
    Returns:
        Boolean True if value is interned and False if it's not
    """
    # We need to check if the value is bytes or string type
    # because we currently can't tell if another object is interned
    if not isinstance(value, (str, bytes)):
        return False

    # In the byteobject.c implmentation only bytes strings
    # that have a size of 1 character are interned
    if isinstance(value, (bytes)) and len(value) < 2:
        return True

    return is_string_interned(value)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os.path
import ctypes
from contrast_vendor import structlog as logging
from contrast.agent import agent_lib

from contrast.agent.settings import Settings
from contrast.utils.loggers import DEFAULT_LOG_PATH, DEFAULT_LOG_LEVEL
from contrast.utils.decorators import fail_loudly


logger = logging.getLogger("contrast")


def init_with_options(enable_logging: bool, log_dir: str, log_level: str):
    """
    Python translation of `init_with_options` + error handling
    """
    c_enable_logging = ctypes.c_bool(enable_logging)
    c_log_dir = ctypes.c_char_p(bytes(log_dir, "utf8"))
    c_log_level = ctypes.c_char_p(bytes(log_level, "utf8"))

    def is_valid_return(code):
        return code == 0

    call(
        agent_lib.modules.LIB_CONTRAST.init_with_options,
        is_valid_return,
        c_enable_logging,
        c_log_dir,
        c_log_level,
    )


def update_log_options(log_level: str):
    valid_return = True

    if not agent_lib.modules.is_initialized():
        return False

    c_enable_logging = ctypes.c_bool(True)
    c_log_level = ctypes.c_char_p(bytes(log_level, "utf8"))

    def is_valid_return(code):
        valid_return = code == 0
        return valid_return

    call(
        agent_lib.modules.LIB_CONTRAST.change_log_settings,
        is_valid_return,
        c_enable_logging,
        c_log_level,
    )

    return valid_return


@fail_loudly("Failed to import agent-lib", return_value=False)
def _import_lib_contrast():
    """
    This import loads the shared object file. For safety (for now), we only want to do
    this if the agent-lib is actually going to be used.

    This should be the only place in the entire agent where we import contrast_agent_lib
    """
    if agent_lib.modules.is_initialized():
        return True

    from contrast_agent_lib import lib_contrast, constants

    agent_lib.modules.LIB_CONTRAST = lib_contrast
    agent_lib.modules.CONSTANTS = constants

    return True


def initialize():
    """
    Import and initialize agent-lib
    """
    logger.debug("Initializing agent-lib")

    if not _import_lib_contrast():
        return False

    agent_lib.modules.LIB_CONTRAST.init_with_options.argtypes = [
        ctypes.c_bool,
        ctypes.c_char_p,
        ctypes.c_char_p,
    ]
    agent_lib.modules.LIB_CONTRAST.init_with_options.restype = ctypes.c_int

    agent_lib.modules.LIB_CONTRAST.change_log_settings.argtypes = [
        ctypes.c_bool,
        ctypes.c_char_p,
    ]
    agent_lib.modules.LIB_CONTRAST.change_log_settings.restype = ctypes.c_int

    init_with_options(
        enable_logging=True,
        log_dir=_get_log_dir(),
        log_level=_get_log_level(),
    )

    agent_lib.initialize_input_tracing()
    agent_lib.initialize_semantic_analysis()

    return True


def call(c_fn, is_valid_return, *args):
    """
    Calls the provided agent-lib function with the specified positional args.

    If the function returns normally, returns the result. Otherwise, retrieves the
    corresponding error message using `last_error_message` and logs profusely.
    @param is_valid_return: function used to handle the return from the agent lib call.
        This was added because different calls to agent lib have different return codes

    This function must remain synchronous - do not use `async def`! Error handling for
    agent-lib functions must occur without another task in the same thread interrupting.
    """
    if not agent_lib.modules.is_initialized():
        logger.error("Calling agent lib without initializing it first")

    fn_result = c_fn(*args)

    if is_valid_return(fn_result):
        return fn_result

    logger.debug(
        "Error from agent-lib - will retrieve error message. "
        "function: %s; args: %s; return: %s",
        c_fn.__name__,
        [getattr(a, "value", a) for a in args],
        fn_result,
    )

    # +1 prevents off-by-1 edge cases
    message_length = agent_lib.modules.LIB_CONTRAST.last_error_message_length() + 1
    stack_length = agent_lib.modules.LIB_CONTRAST.last_error_stack_length() + 1

    c_message_length = ctypes.c_int(message_length)
    c_stack_length = ctypes.c_int(stack_length)

    c_message_buffer = ctypes.create_string_buffer(message_length)
    c_stack_buffer = ctypes.create_string_buffer(stack_length)

    result = agent_lib.modules.LIB_CONTRAST.last_error_message(
        c_message_buffer,
        c_message_length,
        c_stack_buffer,
        c_stack_length,
    )

    if result >= 0:
        logger.debug(
            "Error message from agent-lib: %s; stack: %s",
            c_message_buffer.value.decode("utf8"),
            c_stack_buffer.value.decode("utf8") or "<no stack info>",
        )
    else:
        logger.debug(
            "Failed to retrieve agent-lib last error message. "
            "last_error_message return: %s",
            result,
        )

    return fn_result


def _get_log_dir():
    """
    For now, use the same directory as the agent logger specified in the local config.
    """
    agent_log_dir = Settings().config.get("agent.logger.path") or DEFAULT_LOG_PATH
    # works even for STDOUT because of `dirname`
    return os.path.abspath(os.path.dirname(agent_log_dir))


def _get_log_level():
    return Settings().config.get("agent.logger.level", "").upper() or DEFAULT_LOG_LEVEL



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
agent_lib.initialize() will initialize the module references here. This guarantees that
we handle the `contrast_agent_lib` import safely by restricting it to that function.
"""
LIB_CONTRAST = None
CONSTANTS = None


def is_initialized():
    return LIB_CONTRAST is not None



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re


class DefaultSqlScanner(object):
    """
    Base SQL Scanner for SQL injection rule

    Scans queries for possible attack vectors

    Returns boundaries in string of the vector if one is present
    """

    # States
    STATE_INSIDE_TOKEN = "STATE_INSIDE_TOKEN"
    STATE_INSIDE_NUMBER = "STATE_INSIDE_NUMBER"
    STATE_EXPECTING_TOKEN = "STATE_EXPECTING_TOKEN"
    STATE_INSIDE_DOUBLEQUOTE = "STATE_INSIDE_DOUBLEQUOTE"
    STATE_INSIDE_SINGLEQUOTE = "STATE_INSIDE_SINGLEQUOTE"
    STATE_INSIDE_STRING_ESCAPE_BLOCK = "STATE_INSIDE_STRING_ESCAPE_BLOCK"
    STATE_INSIDE_LINE_COMMENT = "STATE_INSIDE_LINE_COMMENT"  # inside comment that will continue to the end of the line
    STATE_INSIDE_BLOCK_COMMENT = "STATE_INSIDE_BLOCK_COMMENT"  # inside a comment that will end with a closing tag
    STATE_SKIP_NEXT_CHARACTER = "STATE_SKIP_NEXT_CHARACTER"

    # Regex
    DIGIT_REGEX = re.compile(r"\d")
    WHITESPACE_REGEX = re.compile(r"[\s]")
    NON_WHITESPACE_REGEX = re.compile(r"[^\s]")

    OPERATOR_REGEX = re.compile(r"[+=*^/%><!-]")

    # Special characters
    ASTRIX_CHAR = "*"
    DASH_CHAR = "-"
    SLASH_CHAR = "/"

    NO_BOUNDARY = -1

    def __init__(self):
        self._token_boundaries = None

    def crosses_boundary(self, query, index, input_string):
        """
        Checks if the input string crosses the boundary in the query
        """
        last_boundary = 0

        for boundary in self.token_boundaries(query):
            if boundary > index:
                if boundary < index + len(input_string):
                    return last_boundary, boundary

                break

            last_boundary = boundary

        return self.NO_BOUNDARY, self.NO_BOUNDARY

    def token_boundaries(self, query):
        """
        Memoization of token boundaries
        """
        if self._token_boundaries is None:
            self._token_boundaries = self.scan_token_boundaries(query)

        return self._token_boundaries

    def scan_token_boundaries(self, query):
        """
        Scans the query to identify boundaries
        """
        boundaries = []

        if not query:
            return boundaries

        state = DefaultSqlScanner.STATE_EXPECTING_TOKEN
        index = 0

        while index < len(query):
            char = query[index]

            previous_state = state

            state = self.process_state(boundaries, state, char, index, query)

            if state == DefaultSqlScanner.STATE_SKIP_NEXT_CHARACTER:
                index += 1
                state = previous_state
            elif state == DefaultSqlScanner.STATE_INSIDE_STRING_ESCAPE_BLOCK:
                index = self.find_escape_sequence_boundary(query, index + 1)
                state = previous_state
            elif state == DefaultSqlScanner.STATE_INSIDE_BLOCK_COMMENT:
                index = self.find_block_comment_boundary(query, index + 2)
                index += 1
                state = previous_state

                boundaries.append(index)
            elif state == DefaultSqlScanner.STATE_INSIDE_LINE_COMMENT:
                index = self.find_new_line_boundary(query, index + 1)
                state = previous_state

                boundaries.append(index)

            index += 1

        return boundaries

    def process_state(self, boundaries, current_state, char, index, query):
        """
        Processes the current state within the query
        """
        result = None

        if current_state == DefaultSqlScanner.STATE_EXPECTING_TOKEN:
            result = self.process_expecting_token(boundaries, char, index, query)
        elif current_state == DefaultSqlScanner.STATE_INSIDE_NUMBER:
            result = self.process_number(boundaries, char, index)
        elif current_state == DefaultSqlScanner.STATE_INSIDE_TOKEN:
            result = self.process_inside_token(boundaries, char, index, query)
        elif current_state == DefaultSqlScanner.STATE_INSIDE_DOUBLEQUOTE:
            result = self.process_double_quote(boundaries, char, index, query)
        elif current_state == DefaultSqlScanner.STATE_INSIDE_SINGLEQUOTE:
            result = self.process_single_quote(boundaries, char, index, query)

        return result

    def process_expecting_token(self, boundaries, char, index, query):
        """
        Processes the character based on the expected token
        """
        result = DefaultSqlScanner.STATE_EXPECTING_TOKEN

        if char == "'":  # single quote
            boundaries.append(index)
            result = DefaultSqlScanner.STATE_INSIDE_SINGLEQUOTE
        elif char == '"':  # double quote
            boundaries.append(index)
            result = DefaultSqlScanner.STATE_INSIDE_DOUBLEQUOTE
        elif re.match(DefaultSqlScanner.DIGIT_REGEX, char):  # digit
            boundaries.append(index)
            result = DefaultSqlScanner.STATE_INSIDE_NUMBER
        elif self.start_line_comment(char, index, query):  # line comment
            boundaries.append(index)
            result = DefaultSqlScanner.STATE_INSIDE_LINE_COMMENT
        elif self.start_block_comment(char, index, query):  # block comment
            boundaries.append(index)
            result = DefaultSqlScanner.STATE_INSIDE_BLOCK_COMMENT
        elif re.match(DefaultSqlScanner.NON_WHITESPACE_REGEX, char):  # non whitespace
            boundaries.append(index)
            result = DefaultSqlScanner.STATE_INSIDE_TOKEN

        return result

    def process_inside_token(self, boundaries, char, index, query):
        """
        Process the character and returns the new state
        """
        result = DefaultSqlScanner.STATE_INSIDE_TOKEN

        if char == "'":  # single quote
            boundaries.append(index)
            result = DefaultSqlScanner.STATE_INSIDE_SINGLEQUOTE
        elif char == '"':  # double quote
            boundaries.append(index)
            result = DefaultSqlScanner.STATE_INSIDE_DOUBLEQUOTE
        elif self.start_line_comment(char, index, query):  # line comment
            boundaries.append(index)
            result = DefaultSqlScanner.STATE_INSIDE_LINE_COMMENT
        elif self.start_block_comment(char, index, query):  # block comment
            boundaries.append(index)
            result = DefaultSqlScanner.STATE_INSIDE_BLOCK_COMMENT
        elif self.is_operator(char):  # operator
            boundaries.append(index)
            result = DefaultSqlScanner.STATE_EXPECTING_TOKEN
        elif re.match(DefaultSqlScanner.WHITESPACE_REGEX, char):  # whitespace
            boundaries.append(index)
            result = DefaultSqlScanner.STATE_EXPECTING_TOKEN

        return result

    def process_number(self, boundaries, char, index):
        """
        Check if the character is a number
        """
        if re.match(DefaultSqlScanner.DIGIT_REGEX, char) or char == ".":
            return DefaultSqlScanner.STATE_INSIDE_NUMBER

        boundaries.append(index)
        return DefaultSqlScanner.STATE_EXPECTING_TOKEN

    def process_double_quote(self, boundaries, char, index, query):
        """
        Processes the double quote within a query
        """
        if self.escape_char(char):
            return DefaultSqlScanner.STATE_SKIP_NEXT_CHARACTER

        if self.escape_sequence_start(char):
            return DefaultSqlScanner.STATE_INSIDE_STRING_ESCAPE_BLOCK

        if char == '"':
            if self.double_quote_escape_in_double_quote() and self.is_double_quote(
                query, index + 1
            ):
                return DefaultSqlScanner.STATE_SKIP_NEXT_CHARACTER

            boundaries.append(index)
            return DefaultSqlScanner.STATE_EXPECTING_TOKEN

        return DefaultSqlScanner.STATE_INSIDE_DOUBLEQUOTE

    def process_single_quote(self, boundaries, char, index, query):
        """
        Processes a single quote within a query
        """
        if self.escape_char(char):
            return DefaultSqlScanner.STATE_SKIP_NEXT_CHARACTER

        if self.escape_sequence_start(char):
            return DefaultSqlScanner.STATE_INSIDE_STRING_ESCAPE_BLOCK

        if char == "'":
            if self.single_quote_escape_in_single_quote() and self.is_single_quote(
                query, index + 1
            ):
                return DefaultSqlScanner.STATE_SKIP_NEXT_CHARACTER

            boundaries.append(index)
            return DefaultSqlScanner.STATE_EXPECTING_TOKEN

        return DefaultSqlScanner.STATE_INSIDE_SINGLEQUOTE

    def is_double_quote(self, query, index):
        """
        Checks if the character at the index is a double quote
        """
        if index < 0 or index >= len(query):
            return False

        return query[index] == '"'

    def is_single_quote(self, query, index):
        """
        Checks if the character at the index is a single quote
        """
        if index < 0 or index >= len(query):
            return False

        return query[index] == "'"

    def find_escape_sequence_boundary(self, query, index):
        """
        Finds a boundary of an escape sequence
        """
        pos = index

        while pos < len(query):
            char = query[pos]

            if self.escape_sequence_end(char):
                break
            pos += 1

        return pos

    def find_block_comment_boundary(self, query, index):
        """
        Finds a boundary for a block comment
        """
        pos = index

        while pos < len(query):
            char = query[pos]

            if self.end_block_comment(char, pos, query):
                break
            pos += 1

        return pos

    def find_new_line_boundary(self, query, index):
        """
        Finds a new line boundary
        """
        pos = index

        while pos < len(query):
            char = query[pos]

            if char in ["\n", "\r"]:
                break
            pos += 1

        return pos

    def is_operator(self, char):
        """
        Checks if a character is an operator via regex
        """
        return re.match(DefaultSqlScanner.OPERATOR_REGEX, char) is not None

    def start_line_comment(self, char, index, query):
        """
        Validates char is the start of a line comment
        """
        if char != DefaultSqlScanner.DASH_CHAR or not self.query_longer_than_index(
            query, index, 2
        ):
            return False

        return query[index + 1] == DefaultSqlScanner.DASH_CHAR

    def start_block_comment(self, char, index, query):
        """
        Is the current character / sequence of characters the start of a block comment
        We assume '/*' starts the comment by default

        :param char: character to check
        :param index: point in query
        :param query: sql query
        :return: True if start of block comment
        """

        if char != DefaultSqlScanner.SLASH_CHAR or not self.query_longer_than_index(
            query, index, 2
        ):
            return False

        return query[index + 1] == DefaultSqlScanner.ASTRIX_CHAR

    def end_block_comment(self, char, index, query):
        """
        Is the current character / sequence of characters the end of a block comment
        We assume '*/' ends the comment by default

        :param char: character to check
        :param index: point in query
        :param query: sql query
        :return: True if start of block comment
        """
        if char != DefaultSqlScanner.ASTRIX_CHAR or not self.query_longer_than_index(
            query, index, 2
        ):
            return False

        return query[index + 1] == DefaultSqlScanner.SLASH_CHAR

    def double_quote_escape_in_double_quote(self):
        """
        Indicates if '""' inside of double quotes is the equivalent of '\"'
        We assume no by default
        """
        return False

    def single_quote_escape_in_single_quote(self):
        """
        Indicates if "''" inside of single quotes is the equivalent of "\'"
        We assume yes by default
        """
        return True

    def redefines_escape_char(self):
        """
        Does this language let the user redefine the escape character
        We assume no by default
        """
        return False

    def escape_char(self, char):
        """
        Is the character provided an escape character?
        By default, we'll assume
        """
        return char == "\\"

    def support_string_escape_sequence(self):
        """
        Does this language support string escape sequences?
        We assume no by default
        """
        return False

    def escape_sequence_start(self, char):
        """
        Is this the start of a string escape sequence?
        Since escape sequences aren't supported, the answer is always false
        """
        return False

    def escape_sequence_end(self, char):
        """
        Is this the end of a string escape sequence?
        Since escape sequences aren't supported, the answer is always false
        """
        return False

    def query_longer_than_index(self, query, index, diff):
        return len(query) - diff >= index



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import sys

from contrast_vendor.ported_cpython_code.import_functionality import (
    resolve_relative_import_name,
)
from contrast_vendor.wrapt import register_post_import_hook

import builtins
import contrast
from contrast.agent import scope
from contrast.agent.policy import patch_manager, registry
from contrast.agent.assess.policy import trigger_policy
from contrast.agent.settings import Settings
from contrast.reporting import teamserver_messages
from contrast.utils.decorators import fail_quietly
from contrast.utils.library_reader.patched_state import DistributionContext
from contrast.utils.library_reader.utils import (
    normalize_file_name,
    get_file_from_module,
)
from contrast.utils.patch_utils import build_and_apply_patch, wrap_and_watermark


def names_to_possible_sys_module_entries(possible_sys_module_entries):
    """This function converts something like : name = "a.b.c", fromlist = ("d") into
    sys_module_cache = {
        "a": sys.modules["a"],
        "a.b": sys.modules["a.b"],
        "a.b.c": sys.modules["a.b.c"]
        "a.b.c.d": sys.modules["a.b.c.d"] (NOTE: item "d" in the fromlist can be a class or a variable to import.
            We aren't 100% sure if its a module/package because we don't have direct control over what gets set
            on sys.modules. In this scenario this is in part an estimation. Hence the checks in sys.modules
            using sys_module_cache before and after import)
    }
    Calling this function before and after import is used to see what files where actually loaded and then cached.
    This has to be done because __import__ only returns the top level module (e.g sys_modules["a"]) or
    a single specific module if a fromlist is not none.

    arguments: possible_sys_module_entries - a list of possible entries in sys modules
    return: a dictionary containing entries in sys modules based on the tuple possible_sys_module_entries
    """
    sys_module_cache = {}

    if not possible_sys_module_entries:
        return None

    for entry in possible_sys_module_entries:
        sys_module_cache[entry] = sys.modules.get(entry, None)

    return sys_module_cache


def build_possible_sys_module_keys(name, level, global_namespace, fromlist):
    """
    The purpose of this function is to create a list of keys that could be cached in sys.modules
    arguments:
    name - name of module to be imported
    fromlist - list of modules, variables and classes to import
    e.g if the import was performed: from module import SomeClass, fromlist=('SomeClass',)
    """
    possible_sys_module_entries = set()

    if level > 0:
        name = resolve_relative_import_name(name, level, global_namespace)

    if not name:
        return None

    parents = name.split(".")

    for i in range(len(parents)):
        possible_sys_module_entries.add(".".join(parents[0 : i + 1]))

    if fromlist and fromlist[0] != "*":
        for import_item in fromlist:
            possible_sys_module_entries.add(f"{name}.{import_item}")

    return possible_sys_module_entries


def build_new_loaded_files(before_import_sys_modules, after_import_sys_modules):
    """
    The purpose of this function is to compare the values in both dictionaries.
    If before_import_sys_modules doesn't contain a value given a key and
    after_import_sys_modules does, than that module was just loaded and we report on it.
    """
    dist_ctx = DistributionContext()
    modules_loaded = {}

    if not before_import_sys_modules or not after_import_sys_modules:
        return modules_loaded

    for sys_module_key in before_import_sys_modules.keys():
        if (
            not before_import_sys_modules[sys_module_key]
            and after_import_sys_modules[sys_module_key]
        ):
            module = after_import_sys_modules[sys_module_key]
            module_file = get_file_from_module(module)

            # If dist_hash is None, that means we didn't detect the module loaded in the current env
            dist_hash = dist_ctx.get_dist_hash_from_file_path(module_file)
            if dist_hash:
                normalized_file_name = normalize_file_name(module_file)
                files_loaded = modules_loaded.get(dist_hash, None)
                if files_loaded:
                    files_loaded.append(normalized_file_name)
                else:
                    modules_loaded[dist_hash] = [normalized_file_name]

    return modules_loaded


# 0 means attempt abs import
LEVEL = 0


@fail_quietly(
    "Failed to determine sys module keys to perform analysis on",
    return_value=(False, None, None),
)
# Need to use original key name for globals/locals to make sure its unpacked properly when in key=value form.
# Arguments can be passed as either a list (e.g __import__(name, globals_dict, locals_dict, ...)) or
# key=value so we need to make sure we account for both cases
# pylint: disable=redefined-builtin
def pre__import__analysis(
    name, globals=None, locals=None, fromlist=(), level=LEVEL, **kwargs
):
    if not name or name.startswith("contrast"):
        return False, None, None

    possible_sys_module_keys = build_possible_sys_module_keys(
        name, level, globals, fromlist
    )

    if not possible_sys_module_keys:
        return False, None, None

    before_import_sys_modules = names_to_possible_sys_module_entries(
        possible_sys_module_keys
    )

    return True, possible_sys_module_keys, before_import_sys_modules


@fail_quietly("Failed to analyze assess")
def analyze_assess(method_name, result, args, kwargs):
    if scope.in_trigger_scope() or scope.in_contrast_scope():
        return

    trigger_rule = registry.get_triggers_by_rule("unsafe-code-execution")

    trigger_nodes = trigger_rule.find_trigger_nodes("importlib", method_name)

    trigger_policy.apply(trigger_rule, trigger_nodes, result, args, kwargs)


@fail_quietly("Failed to determine loaded files")
def post__import__analysis(
    req_context, before_import_sys_modules, possible_sys_module_keys
):
    after_import_sys_modules = names_to_possible_sys_module_entries(
        possible_sys_module_keys
    )

    loaded_files_dict = build_new_loaded_files(
        before_import_sys_modules, after_import_sys_modules
    )

    if len(loaded_files_dict) > 0:
        if req_context.observed_library_usage is None:
            req_context.observed_library_usage = teamserver_messages.LibraryUsage()

        msg = req_context.observed_library_usage
        for dist_hash, loaded_files in loaded_files_dict.items():
            if len(loaded_files) > 0:
                observation = {"id": dist_hash, "names": loaded_files}
                msg.body["observations"].append(observation)


def build_import_patch(orig_func, _):
    def import_patch(wrapped, instance, args, kwargs):
        del instance

        context = contrast.CS__CONTEXT_TRACKER.current()
        settings = Settings()

        perform_analysis = False
        possible_sys_module_keys = None
        before_import_sys_modules = None
        result = None

        # We only support reporting this information in the context of a request.
        # Verify we can get request context before doing any processing
        if context is not None and settings.is_analyze_libs_enabled() and args:
            with scope.contrast_scope():
                (
                    perform_analysis,
                    possible_sys_module_keys,
                    before_import_sys_modules,
                ) = pre__import__analysis(*args, **kwargs)

        try:
            # Don't perform import in contrast scope because there may be analysis we can
            # do during import (i.e first time executing + caching the module)
            result = wrapped(*args, **kwargs)
        finally:
            if settings.is_assess_enabled():
                analyze_assess("__import__", result, args, kwargs)

        if perform_analysis:
            with scope.contrast_scope():
                post__import__analysis(
                    context, before_import_sys_modules, possible_sys_module_keys
                )

        return result

    return wrap_and_watermark(orig_func, import_patch)


def patch_import(module):
    build_and_apply_patch(module, "__import__", build_import_patch)


def register_patches():
    register_post_import_hook(patch_import, builtins.__name__)
    register_post_import_hook(patch_import, "importlib")


def reverse_patches():
    patch_manager.reverse_patches_by_owner(builtins)

    importlib = sys.modules.get("importlib")
    if importlib is None:  # pragma: no cover
        return

    patch_manager.reverse_patches_by_owner(importlib)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.policy.trigger_actions.default_action import DefaultAction


class FromstringAction(DefaultAction):
    """
    Custom trigger action that implements XXE fromstring rule logic.

    For lxml.etree.fromstring(), if the parser has resolve_entities disabled, then
    `fromstring` is not vulnerable to XXE.
    """

    def is_violated(
        self, source_properties, required_tags, disallowed_tags, orig_args=None
    ) -> bool:
        if not super().is_violated(source_properties, required_tags, disallowed_tags):
            return False

        if orig_args is None or len(orig_args) < 2:
            return True

        parser = orig_args[1]
        if hasattr(parser, "resolve_entities") and parser.resolve_entities is False:
            return False

        return True



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

DEFAULT_LOG_PATH = "contrast-agent.log"
DEFAULT_SECURITY_LOG_PATH = "security.log"
DEFAULT_SECURITY_LOG_LEVEL = "INFO"
DEFAULT_LOG_LEVEL = "INFO"
DEFAULT_PROGNAME = "Contrast Agent"
SECURITY_LOGGER_NAME = "contrast-security-logger"
LOGGER_NAME = "contrast"



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.protect.rule.sql_injection.default_sql_scanner import (
    DefaultSqlScanner,
)


class MysqlSqlScanner(DefaultSqlScanner):
    """
    SQL scanner specifically for MySQL queries
    """

    def start_line_comment(self, char, index, query):
        """
        Valid char is # or subsequent - chars as --
        """
        if char == "#":
            return True

        if char != "-" or not self.query_longer_than_index(query, index, 2):
            return False

        return query[index + 1] == "-"

    def start_block_comment(self, char, index, query):
        """
        Is the current character / sequence of characters the start of a block
        comment. In MySQL, '/*!' is an inline comment that has code that's
        executed, so it does not count as a block comment start
        """

        if char != "/" or not self.query_longer_than_index(query, index, 3):
            return False

        return query[index + 1] == "*" and query[index + 2] != "!"

    def double_quote_escape_in_double_quote(self):
        """
        Indicates if '""' inside of double quotes is the equivalent of '\"'
        """
        return True



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.policy.registry import register_source_nodes
from contrast.agent.policy.utils import CompositeNode

flask_sources = [
    CompositeNode(
        {
            "module": "flask.wrappers",
            "class_name": "Request",
        },
        [
            {
                "method_name": "form",
                "node_type": "MULTIPART_FORM_DATA",
                "tags": ["CROSS_SITE"],
            },
            {
                "method_name": "data",
                "node_type": "BODY",
                "tags": ["CROSS_SITE"],
            },
            {
                "method_name": "cookies",
                "node_type": "COOKIE",
                "tags": ["NO_NEWLINES"],
            },
            {
                "method_name": [
                    "path",
                    "full_path",
                    "base_url",
                    "host",
                    "host_url",
                    "url",
                    "url_root",
                ],
                "node_type": "URI",
                "tags": ["CROSS_SITE"],
            },
        ],
    ),
    CompositeNode(
        {
            "module": "werkzeug.datastructures",
            "class_name": "MultiDict",
        },
        [
            {
                # MultiDict is the storage class for user input parsed from
                # wsgi.input and other sources of user input from environ.
                # Parsing of wsgi.input is deadzoned so we need to track
                # what comes out of this dict.
                # getlist uses the Python dict
                # __getitem__ so we cannot rely on __getitem__ for patching
                "method_name": ["__getitem__", "getlist"],
                "node_type": "PARAMETER",
                "tags": ["CROSS_SITE"],
            },
        ],
    ),
    {
        # Flask uses werkzeug TypeConversionDict as its dict storage class.
        # Cannot patch __getitem__ as it uses Python's dict __getitem__
        "module": "werkzeug.datastructures",
        "class_name": "TypeConversionDict",
        "method_name": "get",
        "node_type": "PARAMETER",
        "tags": ["CROSS_SITE"],
    },
]


register_source_nodes(flask_sources)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.utils.command_scanner import CommandScanner
from contrast.utils.decorators import cached_property


class CommandInjectionREPMixin(object):
    REP_DETECT_PARAMETER_COMMAND_BACKDOORS = "detect_parameter_command_backdoors"
    REP_DETECT_PARAMETER_COMMAND_BACKDOORS_ID = "REP_DETECT_PARAMETER_COMMAND_BACKDOORS"

    REP_DETECT_CHAINED_COMMANDS = "detect_chained_commands"

    @cached_property
    def is_detect_parameter_command_backdoors_enabled(self):
        return self.settings.is_rep_feature_enabled_for_rule(
            self.name, self.REP_DETECT_PARAMETER_COMMAND_BACKDOORS
        )

    @cached_property
    def is_detect_chained_commands_enabled(self):
        return self.settings.is_rep_feature_enabled_for_rule(
            self.name, self.REP_DETECT_CHAINED_COMMANDS
        )

    def get_first_command_chain_index(self, command):
        parsed_results = CommandScanner.parse(command)
        chain_indices = parsed_results.command_chains

        return chain_indices[0] if chain_indices else -1



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.adjusted_span import AdjustedSpan
from contrast.agent.assess.assess_exceptions import ContrastAssessException

BELOW = "BELOW"
LOW_SPAN = "LOW_SPAN"
WITHIN = "WITHIN"
WITHOUT = "WITHOUT"
HIGH_SPAN = "HIGH_SPAN"
ABOVE = "ABOVE"


class Tag(object):
    def __init__(self, length, start_index=0):
        self.start_index = 0
        self.end_index = 0
        self.length = 0

        self.update_range(start_index, start_index + length)

    def __str__(self):
        return f"[{self.start_index},{self.end_index}]"

    def __repr__(self):
        return f"Tag({self.start_index} - {self.end_index}, {self.length})"

    def __len__(self):
        return self.length

    def __eq__(self, other):
        return (
            self.start_index == other.start_index and self.end_index == other.end_index
        )

    def update_range(self, start_index, end_index):
        if start_index < 0:
            raise ContrastAssessException("Unable to set start index negative")

        if end_index < start_index:
            raise ContrastAssessException("Unable to set start index after end index")

        self.start_index = start_index
        self.end_index = end_index

        self.length = end_index - start_index

    def covers(self, index):
        """
        Return true if the tag covers the given position in the string
        """
        return self.start_index <= index < self.end_index

    def range(self):
        # slice is exclusive on the end
        return slice(self.start_index, self.end_index + 1, 1)

    def overlaps(self, other):
        """
        Return True if Tag's overlap in range else False
        :param other: Tag object to compare
        :return: boolean
        """
        if (
            self.start_index < other.start_index <= self.end_index
        ):  # we start below other & end in it
            return True

        if (
            self.start_index >= other.start_index and self.end_index <= other.end_index
        ):  # we start and end in other
            return True

        if (
            self.start_index <= other.end_index < self.end_index
        ):  # we start in other & end above it
            return True

        return False

    def to_span(self, shift=0):
        return AdjustedSpan(self.start_index + shift, self.end_index + shift)

    # These methods shift or set new indexes on the existing Tag object
    def shift(self, index):
        self.update_range(self.start_index + index, self.end_index + index)

    def shift_end(self, index):
        self.update_range(self.start_index, self.end_index + index)

    def set_start(self, start_index):
        self.update_range(start_index, self.end_index)

    def set_end(self, end_index):
        self.update_range(self.start_index, end_index)

    def compare_range(self, start, stop):
        """
        The tag is ______ the range
        rrrrrrr == self.range, the range of the tag
        """
        result = None

        # r starts and stops below
        # rrrrrrrrrrrrr
        #               start       stop
        if self.start_index < start and self.end_index <= start:
            result = BELOW
        # r starts below and finishes within
        # rrrrrrrrrrrrr
        #    start       stop
        if self.start_index < start < self.end_index <= stop:
            result = LOW_SPAN
        # r is between start and stop
        #        rrrrrrrrrrrrrrr
        # start                   stop
        if start <= self.start_index < stop and self.end_index <= stop:
            result = WITHIN
            # r starts below and finishes above stop
        #  rrrrrrrrrrrrrrrrrrrrrrrr
        #     start       stop
        if self.start_index < start and self.end_index > stop:
            result = WITHOUT
            # r starts within and finishes above stop
        #           rrrrrrrrrrrrr
        #   start       stop
        if start <= self.start_index < stop < self.end_index:
            result = HIGH_SPAN
        # starts and stops above
        #                   rrrrrrrrrrrrr
        #  start       stop
        if self.start_index >= stop and self.end_index > stop:
            result = ABOVE

        return result

    def merge(self, other):
        """
        Given a tag, merge its ranges with this one
        such that the lowest start and highest end
        become the values of this tag

        :param other: Tag
        :return: True if the other tag was merged into
        """
        if not self.overlaps(other):
            return False

        start = min(self.start_index, other.start_index)
        finish = max(self.end_index, other.end_index)

        self.update_range(start, finish)

        return True

    def intersect(self, other):
        """
        Given a tag, intersect its ranges with this one
        such that the highest start and lowest end become
        the values of this tag.

        :param other: Tag
        :return: True if the resulting tag has nonzero length
        """
        if not self.overlaps(other):
            return False

        start = max(self.start_index, other.start_index)
        finish = min(self.end_index, other.end_index)

        if start == finish:
            # tags that are directly adjacent no not intersect
            return False

        self.update_range(start, finish)

        return True

    def copy_modified(self, shift):
        """
        Modification to tracked String can change the position and length of the tracked tag

        :param shift: integer value, negative value moves left
        :return: new Tag that has updated indexes
        """
        start = self.start_index + shift

        new_start_index = start if start >= 0 else 0

        new_length = self.length if start >= 0 else (self.length + start)

        return Tag(new_length, new_start_index)



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.

import contrast
from contrast.agent import scope
from contrast.agent.assess.contrast_event import ContrastEvent
from contrast.agent.assess.policy.propagation_node import PropagationNode
from contrast.agent.assess.policy.propagators import (
    PROPAGATOR_ACTIONS,
    STREAM_ACTIONS,
    BasePropagator,
    stream_propagator,
)
from contrast.agent.assess.policy.source_policy import SourceNode, apply_stream_source
from contrast.agent.assess.properties import Properties
from contrast.agent.assess.policy.preshift import Preshift
from contrast.agent.assess.utils import (
    copy_tags_to_offset,
    copy_events,
    copy_from,
    is_tracked,
    get_properties,
    set_properties,
    track_string,
)
from contrast.utils.assess.duck_utils import safe_getattr
from contrast.agent.settings import Settings

from contrast_vendor import structlog as logging

logger = logging.getLogger("contrast")


def apply(nodes, preshift, ret, **kwargs):  # pylint: disable=redefined-builtin
    if not nodes:
        return

    if not scope.in_propagation_scope():
        with scope.propagation_scope():
            for node in nodes:
                if not preshift:
                    continue

                target = node.get_matching_first_target(
                    preshift.obj, ret, preshift.args, preshift.kwargs
                )

                apply_propagator(node, preshift, target, ret)


def track_copy_without_new_event(target, self_obj):
    """
    In general, when a string propagation event results in a copy of the
    original object we track the newly created copy but don't record an event.
    This behavior might be modified in the future.

    Usually, python is smart enough not to create a new string object that is
    simply a copy of the original. However, there are several exceptions to
    this rule.

    This method should only be invoked in these special cases, and they should
    be documented heavily.
    """
    if is_tracked(self_obj):
        track_string(target)
        copy_from(target, self_obj, 0, set())


STREAM_SOURCES = ["read", "read1", "readline", "readlines", "getvalue"]
STREAM_WRITE_METHODS = ["write", "writelines"]


def create_stream_event(node_type, stream, args, kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None or not Settings().is_assess_enabled():
        return None

    module = stream.__class__.__module__
    class_name = stream.__class__.__name__

    if not args and not kwargs:
        # We add an empty string if an empty stream is created for reporting purposes
        args = ("",)

    source_type = safe_getattr(stream, "cs__source_type", None) or "BODY"

    if node_type == "source":
        node = SourceNode(
            module,
            class_name,
            True,
            "__init__",
            "ARG_0,KWARG:initial_value",
            source_type,
        )
    else:
        node = PropagationNode(
            module,
            class_name,
            True,
            "__init__",
            "ARG_0,KWARG:initial_value",
            "RETURN",
            source_type,
            None,
        )

    return ContrastEvent(node, stream, stream, None, args, kwargs, [], 0, None)


def create_stream_source_event(stream, args, kwargs):
    """
    Called directly from C extensions to create source events for __init__
    """
    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None:
        return

    source_type = safe_getattr(stream, "cs__source_type", None) or "BODY"

    init_event = create_stream_event("source", stream, args, kwargs)
    stream.cs__source_event = init_event

    if context.stop_source_creation(context, source_type, None):
        return

    if len(args) > 0:
        properties = get_properties(args[0])
        if properties is not None:
            _add_and_track_event_to_stream(stream, properties, args, kwargs)

    if kwargs and len(kwargs) > 0:
        properties = get_properties(kwargs.get("initial_value"))
        if properties is not None:
            _add_and_track_event_to_stream(stream, properties, args, kwargs)


def _add_and_track_event_to_stream(stream, properties, args, kwargs):
    stream_props = Properties(stream)
    set_properties(stream, stream_props)
    copy_tags_to_offset(stream_props, properties.tags, 0)
    copy_events(stream_props, properties)
    prop_event = create_stream_event("propagation", stream, args, kwargs)
    if prop_event is not None:
        if len(properties.events) > 0:
            prop_event.parent_ids = [properties.events[-1].event_id]
        stream_props.add_event(prop_event)
    stream.cs__tracked = True


def propagate_stream(method_name, target, self_obj, ret, args, kwargs):
    """
    Called directly from C extensions to propagate stream operations.
    """
    args = [] if args is None else args
    kwargs = {} if kwargs is None else kwargs
    preshift = Preshift(self_obj, args, kwargs)

    if method_name in STREAM_WRITE_METHODS:
        stream_propagator.propagate_stream_write(method_name, preshift, target, ret)

    elif (
        self_obj.cs__tracked
        and get_properties(self_obj)
        and get_properties(self_obj).tags
    ):
        propagation_method = STREAM_ACTIONS.get(method_name)
        if propagation_method is None:
            return

        propagation_method(method_name, preshift, target, ret)

    # If the stream is already considered tracked, it will not be treated as a
    # source.
    elif self_obj.cs__source and method_name in STREAM_SOURCES:
        apply_stream_source(method_name, target, self_obj, ret, args, kwargs)


def cast_special_case(method_name, args, target):
    """
    We don't want to record an event when `target is source`.

    This will never happen for bytearrays, because they are always at least copied.
    For bytearrays, we check if the source and target are ==, and if so we copy
    tags over without recording a new event.

    It would be better if there were some way to check for this slightly later,
    based on the policy node's information.

    Currently, every other string function is an instance method, so we short-
    circuit out of propagation if `self_obj is ret`. For casting, self_obj is None
    because casting (which is really __new__ for unicode/bytes) is a module method
    """
    if method_name == "CAST":
        source = args and args[0]
        if target is source:
            return True
        if (
            isinstance(source, bytearray)
            and isinstance(target, bytearray)
            and target == source
        ):
            track_copy_without_new_event(target, source)
            return True
    return False


def translate_special_case(method_name, self_obj, ret):
    """
    py27 unicode objects and py3x str objects returned by str.translate()
    are always new objects, so in order to keep our behavior consistent we
    do not create a new event in the case where new == original

    If this method returns True, we should copy all tags from
    self_obj to target without creating a new event.
    """
    return (
        method_name == "translate"
        and ret == self_obj
        and (
            (PY2 and isinstance(self_obj, unicode))
            or (PY3 and isinstance(self_obj, str))
        )
    )


def apply_propagator(propagator_node, preshift, target, ret):
    if not propagator_node or not target:
        return

    if isinstance(target, dict):
        for key, value in target.items():
            apply_propagator(propagator_node, preshift, key, ret)
            apply_propagator(propagator_node, preshift, value, ret)
    else:
        propagate_string(propagator_node, preshift, target, ret)


def propagate_string(propagator_node, preshift, target, ret):
    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None:
        return

    if context.stop_propagation:
        return

    action = propagator_node.action
    propagator_class = PROPAGATOR_ACTIONS.get(action, BasePropagator)

    propagator = propagator_class(propagator_node, preshift, target)

    if not propagator.needs_propagation:
        return

    propagator.track_and_propagate(ret)
    logger.debug(
        "Propagator %s found: propagated to %s", propagator_node.name, id(target)
    )
    context.propagated()



# -*- coding: utf-8 -*-
# Copyright © 2023 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.rules.dataflow_rule import DataflowRule
from contrast.agent.policy.registry import register_trigger_rule
from contrast.agent.policy.utils import CompositeNode


DISALLOWED_TAGS = [
    "CUSTOM_ENCODED_NOSQL_INJECTION",
    "CUSTOM_ENCODED",
    "CUSTOM_VALIDATED_NOSQL_INJECTION",
    "CUSTOM_VALIDATED",
    "LIMITED_CHARS",
    "JS_ENCODED",
]


nosql_triggers = [
    CompositeNode(
        {
            "module": "pymongo.collection",
            "protect_mode": True,
        },
        [
            {
                "class_name": "Collection",
                "method_name": "find",
                "source": "ARG_0,KWARG:filter",
            },
            {
                "module": "pymongo.collection",
                "class_name": "Collection",
                "method_name": "insert_one",
                "source": "ARG_0,KWARG:document",
            },
            {
                "module": "pymongo.collection",
                "class_name": "Collection",
                "method_name": "insert_many",
                "source": "ARG_0,KWARG:documents",
            },
            {
                "module": "pymongo.collection",
                "class_name": "Collection",
                "method_name": "_delete_retryable",
                "source": "ARG_0,KWARG:criteria",
            },
            {
                "module": "pymongo.collection",
                "class_name": "Collection",
                "method_name": "_update_retryable",
                "source": "ARG_1,KWARG:document",
            },
        ],
    ),
]


register_trigger_rule(
    DataflowRule.from_nodes(
        "nosql-injection",
        nosql_triggers,
        disallowed_tags=DISALLOWED_TAGS,
    )
)
